# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from enum import Enum
from msprobe.visualization.utils import GraphConst, ToolTip
from msprobe.core.common.const import CompareConst

SUMMARY_DESCRIPTION = "此节点所有输入输出的统计量误差, 颜色越深代表调试侧与标杆侧的偏差越大"
REAL_DATA_DESCRIPTION = "此节点所有输入输出的 tensor 误差, 颜色越深代表调试侧与标杆侧的偏差越大"
MD5_DESCRIPTION_N = "与标杆相比, 此节点任意输入输出的md5值不同"
MD5_DESCRIPTION_Y = "与标杆相比, 此节点所有输入输出的md5值相同"
NOT_MATCHED = "比对过程中节点未匹配上"


class NodeColors(Enum):
    # 枚举值后缀数字越小, 颜色越浅
    # value值左闭右开, 两个值相同代表固定值
    YELLOW_1 = ("#FFFCF3", {
        GraphConst.SUMMARY_COMPARE: {GraphConst.VALUE: [0, 0.3], GraphConst.DESCRIPTION: SUMMARY_DESCRIPTION,
                                     GraphConst.ACCURACY_LEVEL: CompareConst.PASS},
        GraphConst.REAL_DATA_COMPARE: {GraphConst.VALUE: [0, 0.3], GraphConst.DESCRIPTION: REAL_DATA_DESCRIPTION,
                                       GraphConst.ACCURACY_LEVEL: CompareConst.PASS},
        GraphConst.MD5_COMPARE: {GraphConst.VALUE: [0, 0.3], GraphConst.DESCRIPTION: MD5_DESCRIPTION_Y,
                                 GraphConst.ACCURACY_LEVEL: CompareConst.PASS},
    })
    ORANGE_1 = ("#FFDC7F", {
        GraphConst.SUMMARY_COMPARE: {GraphConst.VALUE: [0.3, 0.6], GraphConst.DESCRIPTION: SUMMARY_DESCRIPTION,
                                     GraphConst.ACCURACY_LEVEL: CompareConst.WARNING},
        GraphConst.REAL_DATA_COMPARE: {GraphConst.VALUE: [0.3, 0.6], GraphConst.DESCRIPTION: REAL_DATA_DESCRIPTION,
                                       GraphConst.ACCURACY_LEVEL: CompareConst.WARNING},
        GraphConst.MD5_COMPARE: {GraphConst.VALUE: [0.3, 0.6], GraphConst.DESCRIPTION: MD5_DESCRIPTION_N,
                                 GraphConst.ACCURACY_LEVEL: CompareConst.WARNING},
    })
    RED = ("#FF704D", {
        GraphConst.SUMMARY_COMPARE: {GraphConst.VALUE: [0.6, 1], GraphConst.DESCRIPTION: SUMMARY_DESCRIPTION,
                                     GraphConst.ACCURACY_LEVEL: CompareConst.ERROR},
        GraphConst.REAL_DATA_COMPARE: {GraphConst.VALUE: [0.6, 1], GraphConst.DESCRIPTION: REAL_DATA_DESCRIPTION,
                                       GraphConst.ACCURACY_LEVEL: CompareConst.ERROR},
        GraphConst.MD5_COMPARE: {GraphConst.VALUE: [0.6, 1], GraphConst.DESCRIPTION: MD5_DESCRIPTION_N,
                                 GraphConst.ACCURACY_LEVEL: CompareConst.ERROR},
    })
    GREY = ("#C7C7C7", {
        GraphConst.VALUE: [], GraphConst.DESCRIPTION: NOT_MATCHED, GraphConst.ACCURACY_LEVEL: GraphConst.UNMATCHED
    })

    def __init__(self, hex_value, mode_info):
        self.hex_value = hex_value
        self.mode_info = mode_info

    @staticmethod
    def get_node_colors(mode):
        """
        获取不同比对模式下的颜色说明
        Args:
            mode: 比对模式
        Returns: 颜色说明
        """
        return {
            color.hex_value: color.get_info_by_mode(mode) for color in NodeColors if color.get_info_by_mode(mode)
        }

    @staticmethod
    def get_node_error_status(mode, value):
        """
        判断精度数据比对指标是否大于基准值
        Args:
            mode: 比对模式
            value: 精度数据比对指标
        Returns: bool
        """
        info = NodeColors.ORANGE_1.get_info_by_mode(mode)
        if info and GraphConst.VALUE in info:
            value_range = info[GraphConst.VALUE]
            return value > value_range[0]
        return False

    def get_info_by_mode(self, mode):
        if isinstance(self.mode_info, dict):
            # 检查是否是模式特定的信息
            if isinstance(next(iter(self.mode_info.values())), dict):
                return self.mode_info.get(mode, {})
            else:
                # 所有模式共享相同的信息
                return self.mode_info
        return {}



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import os
from dataclasses import dataclass
from typing import Any, Optional
from tqdm import tqdm

from msprobe.core.common.const import Const, CompareConst
from msprobe.core.common.file_utils import FileOpen, create_directory, write_csv, load_json, load_yaml
from msprobe.core.common.utils import add_time_as_suffix
from msprobe.mindspore.api_accuracy_checker.api_info import ApiInfo
from msprobe.mindspore.api_accuracy_checker.api_runner import api_runner, ApiInputAggregation
from msprobe.mindspore.api_accuracy_checker.base_compare_algorithm import compare_algorithms
from msprobe.mindspore.api_accuracy_checker.data_manager import DataManager
from msprobe.mindspore.api_accuracy_checker.utils import (check_and_get_from_json_dict, global_context,
                                                          trim_output_compute_element_list)
from msprobe.mindspore.common.const import MsCompareConst
from msprobe.mindspore.common.log import logger
from msprobe.mindspore.api_accuracy_checker import torch_mindtorch_importer
from msprobe.core.dump.data_dump.data_collector import build_data_collector
from msprobe.core.common.utils import Const, print_tools_ends_info, DumpPathAggregation
from msprobe.core.dump.data_dump.data_processor.base import ModuleForwardInputsOutputs, ModuleBackwardInputsOutputs

cur_path = os.path.dirname(os.path.realpath(__file__))
yaml_path = os.path.join(cur_path, MsCompareConst.SUPPORTED_API_LIST_FILE)


class BasicInfoAndStatus:
    def __init__(self, api_name, bench_dtype, tested_dtype, shape, status, err_msg) -> None:
        self.api_name = api_name
        self.bench_dtype = bench_dtype
        self.tested_dtype = tested_dtype
        self.shape = shape
        self.status = status
        self.err_msg = err_msg


class ResultCsvEntry:
    def __init__(self) -> None:
        self.forward_pass_status = None
        self.backward_pass_status = None
        self.forward_err_msg = ""
        self.backward_err_msg = ""
        self.overall_err_msg = None


class ProcessResultPacket:
    def __init__(self, process_status, result, err_msg) -> None:
        self.process_status = process_status
        self.result = result
        self.err_msg = err_msg


@dataclass
class Config:
    execution_mode: str
    dump_path: str
    task: str
    level: str
    scope: Optional[Any]
    list: Optional[Any]
    framework: str
    data_mode: str
    file_format: str
    dump_tensor_data_dir: str
    async_dump: bool
    summary_mode: Optional[Any] = None


class ApiAccuracyChecker:
    def __init__(self, args):
        self.api_infos = dict()
        self.data_manager = DataManager(args.out_path, args.result_csv_path)  # 在初始化时实例化 DataManager
        self.save_error_data = args.save_error_data
        if self.save_error_data:
            config, dump_path_aggregation = self.init_save_error_data(args)
            self.data_collector = build_data_collector(config)
            self.data_collector.update_dump_paths(dump_path_aggregation)

    @staticmethod
    def init_save_error_data(args):
        config = Config(
            execution_mode="pynative",
            dump_path=f"{args.out_path}",
            dump_tensor_data_dir=f"{args.out_path}",
            task="tensor",  # 任务类型,模拟保存tensor数据
            level="L1",  # 级别
            scope=None,  # 作用域 (None)
            list=None,  # API 列表 (None)
            framework=Const.MS_FRAMEWORK,  # 框架类型
            data_mode="all",
            file_format="npy",
            async_dump=False
        )

        dump_dir = f"{args.out_path}"
        dump_data_dir = os.path.join(dump_dir, "error_data")
        create_directory(dump_data_dir)
        dump_path_aggregation = DumpPathAggregation()
        dump_path_aggregation.dump_file_path = os.path.join(dump_dir, "dump.json")
        dump_path_aggregation.stack_file_path = os.path.join(dump_dir, "stack.json")
        dump_path_aggregation.dump_error_info_path = os.path.join(dump_dir, "dump_error_info.log")
        dump_path_aggregation.dump_tensor_data_dir = dump_data_dir
        return config, dump_path_aggregation

    @staticmethod
    def prepare_api_input_aggregation(api_info, forward_or_backward=Const.FORWARD):
        """
        Args:
            api_info: ApiInfo
            forward_or_backward: str
        Returns:
            ApiInputAggregation
        """
        forward_inputs = api_info.get_compute_element_list(Const.FORWARD, Const.INPUT)
        kwargs = api_info.get_kwargs()
        if forward_or_backward == Const.FORWARD:
            gradient_inputs = None
        else:
            gradient_inputs = api_info.get_compute_element_list(Const.BACKWARD, Const.INPUT)
        return ApiInputAggregation(forward_inputs, kwargs, gradient_inputs)

    @staticmethod
    def is_api_checkable(api_name_str):
        '''
        Args:
            api_name_str: str, e.g. "MintFunctional.relu.0.forward", key in data field of api_info.json
        Returns:
            is_checkable: bool
        Description:
            tell whether this api is checkable based on the key in "data" dict in api_info.json
        '''
        api_name_str_list = api_name_str.split(Const.SEP)
        if len(api_name_str_list) < MsCompareConst.API_NAME_STR_LENGTH:
            return False
        api_type_str = api_name_str_list[0]
        real_api_str = Const.SEP.join(api_name_str_list[1:-2])
        api_list = load_yaml(yaml_path)
        supported_tensor_api_list = api_list.get(MsCompareConst.SUPPORTED_TENSOR_LIST_KEY)
        supported_fusion_api_list = MsCompareConst.SUPPORTED_FUSION_LIST
        if api_type_str in (MsCompareConst.MINT, MsCompareConst.MINT_FUNCTIONAL) \
                and global_context.get_framework() == Const.MS_FRAMEWORK:
            return True
        if api_type_str in MsCompareConst.MT_VALID_API_TYPES \
                and global_context.get_framework() == Const.MT_FRAMEWORK:
            return True
        if api_type_str == MsCompareConst.TENSOR_API and real_api_str in supported_tensor_api_list \
                and global_context.get_framework() == Const.MS_FRAMEWORK:
            return True
        if api_type_str == MsCompareConst.FUNCTIONAL_API and real_api_str in supported_fusion_api_list \
                and global_context.get_framework() == Const.MS_FRAMEWORK:
            return True
        return False

    def post_forward_hook(self, api_or_module_name, primitive_instance, args, kwargs, output):
        self.data_collector.update_api_or_module_name(api_or_module_name)
        module_input_output = ModuleForwardInputsOutputs(args=args, kwargs=kwargs, output=output)
        self.data_collector.forward_data_collect_only_tensor(
            api_or_module_name,
            primitive_instance,
            os.getpid(),
            module_input_output
        )

    def backward_hook(self, api_or_module_name, module, grad_input, grad_output):
        self.data_collector.update_api_or_module_name(api_or_module_name)

        module_input_output = ModuleBackwardInputsOutputs(grad_input=grad_output, grad_output=grad_input)
        self.data_collector.backward_data_collect_only_tensor(
            api_or_module_name,
            module,
            os.getpid(),
            module_input_output
        )

    def run_and_compare_helper(self, api_info, api_name_str, api_input_aggregation, forward_or_backward):
        """
        Args:
            api_info: ApiInfo
            api_name_str: str
            api_input_aggregation: ApiInputAggregation
            forward_or_backward: str: Union["forward", "backward"]

        Return:
            output_list: List[tuple(str, str, BasicInfoAndStatus, dict{str: CompareResult})]

        Description:
            get mindspore api output, run torch api and get output.
            compare output.
            record compare result.
        """
        # get output
        if global_context.get_is_constructed():
            if forward_or_backward == Const.FORWARD:
                tested_outputs, inputs, kwargs, forward_result_tuple = api_runner(api_input_aggregation, api_name_str,
                                                                                  forward_or_backward,
                                                                                  global_context.get_framework())
            elif forward_or_backward == Const.BACKWARD:
                tested_outputs, gradient_inputs, backward_result_tuple = api_runner(api_input_aggregation, api_name_str,
                                                                                    forward_or_backward,
                                                                                    global_context.get_framework())
            else:
                tested_outputs = api_runner(api_input_aggregation, api_name_str,
                                            forward_or_backward, global_context.get_framework())
        else:
            tested_outputs = api_info.get_compute_element_list(forward_or_backward, Const.OUTPUT)

        bench_outputs = api_runner(api_input_aggregation, api_name_str, forward_or_backward, Const.PT_FRAMEWORK)

        tested_outputs = trim_output_compute_element_list(tested_outputs, forward_or_backward)
        bench_outputs = trim_output_compute_element_list(bench_outputs, forward_or_backward)
        if len(tested_outputs) != len(bench_outputs):
            logger.warning(f"ApiAccuracyChecker.run_and_compare_helper: api: {api_name_str}.{forward_or_backward}, "
                           "number of bench outputs and tested outputs is different, comparing result can be wrong. "
                           f"tested outputs: {len(tested_outputs)}, bench outputs: {len(bench_outputs)}")

        # compare output
        output_list = []
        for i, (bench_out, tested_out) in enumerate(zip(bench_outputs, tested_outputs)):
            api_name_with_slot = Const.SEP.join([api_name_str, forward_or_backward, Const.OUTPUT, str(i)])
            bench_dtype = bench_out.get_dtype()
            tested_dtype = tested_out.get_dtype()
            shape = bench_out.get_shape()

            compare_result_dict = dict()
            for compare_algorithm_name, compare_algorithm in compare_algorithms.items():
                compare_result = compare_algorithm(bench_out, tested_out)
                compare_result_dict[compare_algorithm_name] = compare_result

            if compare_result_dict.get(CompareConst.COSINE).pass_status == CompareConst.PASS and \
                    compare_result_dict.get(CompareConst.MAX_ABS_ERR).pass_status == CompareConst.PASS:
                status = CompareConst.PASS
                err_msg = ""

            else:
                status = CompareConst.ERROR
                err_msg = (compare_result_dict.get(CompareConst.COSINE).err_msg +
                           compare_result_dict.get(CompareConst.MAX_ABS_ERR).err_msg)
                if forward_or_backward == Const.FORWARD and self.save_error_data \
                        and global_context.get_is_constructed():
                    api_name_str_backward = f"{api_name_str}{Const.SEP}{Const.FORWARD}"
                    self.post_forward_hook(api_name_str_backward, None, inputs, kwargs, forward_result_tuple)

                if forward_or_backward == Const.BACKWARD and self.save_error_data \
                        and global_context.get_is_constructed():
                    api_name_str_backward = f"{api_name_str}{Const.SEP}{Const.BACKWARD}"
                    self.backward_hook(api_name_str_backward, None, gradient_inputs, backward_result_tuple)

            basic_info_status = \
                BasicInfoAndStatus(api_name_with_slot, bench_dtype, tested_dtype, shape, status, err_msg)
            output_list.append(tuple([api_name_str, forward_or_backward, basic_info_status, compare_result_dict]))
        return output_list

    def parse(self, api_info_path):

        api_info_dict = load_json(api_info_path)

        # init global context
        task = check_and_get_from_json_dict(api_info_dict, MsCompareConst.TASK_FIELD,
                                            "task field in api_info.json", accepted_type=str,
                                            accepted_value=(MsCompareConst.STATISTICS_TASK,
                                                            MsCompareConst.TENSOR_TASK))
        try:
            framework = check_and_get_from_json_dict(api_info_dict, MsCompareConst.FRAMEWORK,
                                                     "framework field in api_info.json", accepted_type=str,
                                                     accepted_value=(Const.MS_FRAMEWORK,
                                                                     Const.MT_FRAMEWORK))
        except Exception as e:
            framework = Const.MS_FRAMEWORK
            logger.warning(f"JSON parsing error in framework field: {e}")

        if framework == Const.MT_FRAMEWORK and not torch_mindtorch_importer.is_valid_pt_mt_env:
            raise Exception(f"Please check if you have a valid PyTorch and MindTorch environment")

        is_constructed = task == MsCompareConst.STATISTICS_TASK
        if not is_constructed:
            dump_data_dir = check_and_get_from_json_dict(api_info_dict, MsCompareConst.DUMP_DATA_DIR_FIELD,
                                                         "dump_data_dir field in api_info.json", accepted_type=str)
        else:
            dump_data_dir = ""
        global_context.init(is_constructed, dump_data_dir, framework)

        api_info_data = check_and_get_from_json_dict(api_info_dict, MsCompareConst.DATA_FIELD,
                                                     "data field in api_info.json", accepted_type=dict)
        for api_name, api_info in api_info_data.items():
            if not self.is_api_checkable(api_name):
                continue
            forbackward_str = api_name.split(Const.SEP)[-1]
            if forbackward_str not in (Const.FORWARD, Const.BACKWARD):
                logger.warning(f"api: {api_name} is not recognized as forward api or backward api, skip this.")
            api_name = Const.SEP.join(api_name.split(Const.SEP)[:-1])  # www.xxx.yyy.zzz --> www.xxx.yyy
            if api_name not in self.api_infos:
                self.api_infos[api_name] = ApiInfo(api_name)

            if forbackward_str == Const.FORWARD:
                self.api_infos[api_name].load_forward_info(api_info)
            else:
                self.api_infos[api_name].load_backward_info(api_info)

    def process_forward(self, api_name_str, api_info):
        """处理前向检查"""
        if not api_info.check_forward_info():
            logger.debug(f"api: {api_name_str} is lack of forward information, skip forward check.")
            process_result_packet = ProcessResultPacket(process_status=MsCompareConst.ProcessStatus.API_NOT_FOUND,
                                                        result=None,
                                                        err_msg=f"forward info of {api_name_str} is not found")
            return process_result_packet

        try:
            forward_inputs_aggregation = self.prepare_api_input_aggregation(api_info, Const.FORWARD)
        except Exception as e:
            logger.warning(f"Exception occurs when getting inputs for {api_name_str} forward api. "
                           f"Skipping forward check. Detailed exception information: {e}.")
            process_result_packet = ProcessResultPacket(process_status=MsCompareConst.ProcessStatus.EXCEPTION_SKIP,
                                                        result=None, err_msg=f"{e}")
            return process_result_packet

        try:
            forward_output_list = self.run_and_compare_helper(api_info, api_name_str, forward_inputs_aggregation,
                                                              Const.FORWARD)
        except Exception as e:
            logger.warning(f"Exception occurs when running and comparing {api_name_str} forward api. "
                           f"Detailed exception information: {e}.")
            process_result_packet = ProcessResultPacket(process_status=MsCompareConst.ProcessStatus.EXCEPTION_SKIP,
                                                        result=None, err_msg=f"{e}")
            return process_result_packet

        process_result_packet = ProcessResultPacket(process_status=MsCompareConst.ProcessStatus.SUCCESS,
                                                    result=forward_output_list, err_msg="")
        return process_result_packet

    def process_backward(self, api_name_str, api_info):
        """处理反向检查"""
        if not api_info.check_backward_info():
            logger.debug(f"api: {api_name_str} is lack of backward information, skipping backward check.")
            process_result_packet = ProcessResultPacket(process_status=MsCompareConst.ProcessStatus.API_NOT_FOUND,
                                                        result=None,
                                                        err_msg=f"backward info of {api_name_str} is not found")
            return process_result_packet

        try:
            backward_inputs_aggregation = self.prepare_api_input_aggregation(api_info, Const.BACKWARD)
        except Exception as e:
            logger.warning(f"Exception occurs when getting inputs for {api_name_str} backward api. "
                           f"Skipping backward check. Detailed exception information: {e}.")
            process_result_packet = ProcessResultPacket(process_status=MsCompareConst.ProcessStatus.EXCEPTION_SKIP,
                                                        result=None, err_msg=f"{e}")
            return process_result_packet

        try:
            backward_output_list = self.run_and_compare_helper(api_info, api_name_str, backward_inputs_aggregation,
                                                               Const.BACKWARD)
        except Exception as e:
            logger.warning(f"Exception occurs when running and comparing {api_name_str} backward api. "
                           f"Detailed exception information: {e}.")
            process_result_packet = ProcessResultPacket(process_status=MsCompareConst.ProcessStatus.EXCEPTION_SKIP,
                                                        result=None, err_msg=f"{e}")
            return process_result_packet

        process_result_packet = ProcessResultPacket(process_status=MsCompareConst.ProcessStatus.SUCCESS,
                                                    result=backward_output_list, err_msg="")
        return process_result_packet

    def run_and_compare(self):
        for api_name_str, api_info in tqdm(self.api_infos.items()):
            if not self.data_manager.is_unique_api(api_name_str):
                continue

            # 处理前向
            process_result_packet = self.process_forward(api_name_str, api_info)
            if process_result_packet.process_status is MsCompareConst.ProcessStatus.SUCCESS:
                self.data_manager.record(process_result_packet.result)
            elif process_result_packet.process_status == MsCompareConst.ProcessStatus.EXCEPTION_SKIP:
                self.data_manager.record_exception_skip(api_name_str, Const.FORWARD, process_result_packet.err_msg)

            # 处理反向
            process_result_packet = self.process_backward(api_name_str, api_info)
            if process_result_packet.process_status is MsCompareConst.ProcessStatus.SUCCESS:
                self.data_manager.record(process_result_packet.result)
            elif process_result_packet.process_status == MsCompareConst.ProcessStatus.EXCEPTION_SKIP:
                self.data_manager.record_exception_skip(api_name_str, Const.BACKWARD, process_result_packet.err_msg)

            self.data_manager.save_results(api_name_str)


# coding=utf-8
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
Function:
This file mainly involves the CompareError function.
"""


class CompareError(Exception):
    """
    The class for compare error
    """
    MSACCUCMP_NONE_ERROR = 0
    MSACCUCMP_UNKNOWN_ERROR = 1
    MSACCUCMP_NO_DUMP_FILE_ERROR = 2
    MSACCUCMP_INVALID_PATH_ERROR = 3
    MSACCUCMP_PARSE_DUMP_FILE_ERROR = 4
    MSACCUCMP_MALLOC_ERROR = 5
    MSACCUCMP_INVALID_PARAM_ERROR = 6
    MSACCUCMP_OPEN_FILE_ERROR = 7
    MSACCUCMP_CLOSE_FILE_ERROR = 8
    MSACCUCMP_OPEN_DIR_ERROR = 9
    MSACCUCMP_CLOSE_DIR_ERROR = 10
    MSACCUCMP_INVALID_DUMP_DATA_ERROR = 11
    MSACCUCMP_INDEX_OUT_OF_BOUNDS_ERROR = 12
    MSACCUCMP_PARSER_JSON_FILE_ERROR = 13
    MSACCUCMP_MEMSET_ERROR = 14
    MSACCUCMP_LEFT_DUMP_ZERO_ERROR = 15
    MSACCUCMP_RIGHT_DUMP_ZERO_ERROR = 16
    MSACCUCMP_MATCH_MORE_FILE_ERROR = 17
    MSACCUCMP_DUMP_FILE_ERROR = 18
    MSACCUCMP_WRITE_FILE_ERROR = 19
    MSACCUCMP_INVALID_DUMP_TYPE_ERROR = 20
    MSACCUCMP_INVALID_DATA_TYPE_ERROR = 21
    MSACCUCMP_INVALID_FORMAT_ERROR = 22
    MSACCUCMP_INVALID_SHAPE_ERROR = 23
    MSACCUCMP_INVALID_TYPE_ERROR = 24
    MSACCUCMP_INVALID_TARGET_ERROR = 25
    MSACCUCMP_INVALID_CONVERT_FUNC_ERROR = 26
    MSACCUCMP_SHAPE_CONVERT_ERROR = 27
    MSACCUCMP_INVALID_GOLDEN_EVN_ERROR = 28
    MSACCUCMP_INVALID_JSON_FILE_ERROR = 29
    MSACCUCMP_UNSUPPORTED_COMPARE_ERROR = 30
    MSACCUCMP_INVALID_ALGORITHM_ERROR = 31
    MSACCUCMP_DELETE_FILE_ERROR = 32
    MSACCUCMP_INVALID_FILE_ERROR = 33
    MSACCUCMP_INVALID_FRACTAL_NZ_DUMP_DATA_ERROR = 34
    MSACCUCMP_INVALID_OVERFLOW_TYPE_ERROR = 35
    MSACCUCMP_INVALID_OVERFLOW_STATUS_ERROR = 36
    MSACCUCMP_INVALID_TASK_TYPE = 37
    MSACCUCMP_MISSING_FILE = 38
    MSACCUCMP_UNMATCH_STANDARD_DUMP_SIZE = 39
    MSACCUCMP_UNMATCH_DATA_SHAPE_ERROR = 40
    MSACCUCMP_INVALID_SLICE_DATA = 41
    MSACCUCMP_INVALID_INPUT_MAPPING = 42
    MSACCUCMP_FILE_TOO_LARGE_ERROR = 43
    MSACCUCMP_SYMLINK_ERROR = 44
    MSACCUCMP_FILE_EXISTS_ERROR = 45
    MSACCUCMP_DANGER_FILE_ERROR = 46
    MSACCUCMP_PARSE_NANO_DUMP_FILE_ERROR = 47
    MSACCUCMP_NAME_ERROR = 48

    def __init__(self: any, code: int, message: str = "") -> None:
        super().__init__(code)
        self.code = code
        self.message = message



#!/usr/bin/env python
# coding=utf-8
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
Function:
DumpDataConversion class. This class mainly involves the convert_data function.
"""

import os
import sys
import argparse
import numpy as np

from msprobe.msaccucmp.cmp_utils import utils, path_check
from msprobe.msaccucmp.dump_parse import big_dump_data, dump_utils
from msprobe.msaccucmp.cmp_utils import log
from msprobe.msaccucmp.cmp_utils.utils import safe_path_string
from msprobe.msaccucmp.cmp_utils.constant.const_manager import ConstManager
from msprobe.msaccucmp.cmp_utils.multi_process.multi_convert_process import MultiConvertProcess
from msprobe.msaccucmp.cmp_utils.reg_manager import RegManager
from msprobe.msaccucmp.cmp_utils.constant.compare_error import CompareError


class DumpDataConversion:
    """
    The class for dump data conversion
    """

    SUPPORT_TARGET = ('dump', 'numpy')
    SUPPORT_TYPE = ('caffe', 'tf', 'quant', 'offline')

    def __init__(self: any) -> None:
        parse = argparse.ArgumentParser()
        parse.add_argument("-target", dest="target",
                           help="<Required> the target after conversion, "
                                "numpy or dump", required=True)
        parse.add_argument("-type", dest="type",
                           help="<Required> the type for data, "
                                "caffe, tf, quant or offline",
                           required=True)
        parse.add_argument("-i", dest="input_path", type=safe_path_string,
                           help="<Required> the input path", required=True)
        parse.add_argument("-o", dest="output_path", default="", type=safe_path_string,
                           help="<Required> the output path", required=True)

        args, _ = parse.parse_known_args(sys.argv[1:])
        self.target = args.target
        self.input_path = os.path.realpath(args.input_path)
        if os.path.islink(os.path.abspath(args.output_path)):
            log.print_error_log('The path "%r" is a softlink, not permitted.' % args.output_path)
            raise CompareError(CompareError.MSACCUCMP_INVALID_PATH_ERROR)
        self.output_path = os.path.realpath(args.output_path)
        self.type = args.type
        self.multi_process = MultiConvertProcess(self._convert_file, self.input_path, self.output_path)

        self.layer_name_switch = {
            "caffe": self._get_standard_layer_name,
            "tf": self._get_standard_layer_name,
            "quant": self._get_quant_layer_name,
            "offline": self._get_offline_layer_name
        }

        self.file_name_switch = {
            "caffe": lambda name: name + ".pb",
            "tf": lambda name: name + ".pb",
            "quant": lambda name: name + ".quant",
            "offline": lambda name: name
        }

    def check_arguments_valid(self: any) -> int:
        """
        Check arguments valid
        :return: error code
        """
        return_code = path_check.check_path_valid(self.input_path, True)
        if return_code != CompareError.MSACCUCMP_NONE_ERROR:
            return return_code

        return_code = path_check.check_output_path_valid(self.output_path, True)
        if return_code != CompareError.MSACCUCMP_NONE_ERROR:
            return return_code

        if self.target not in self.SUPPORT_TARGET:
            log.print_error_log("The input of target argument does not support, only supports dump and numpy.")
            return CompareError.MSACCUCMP_INVALID_TARGET_ERROR
        if self.type not in self.SUPPORT_TYPE:
            log.print_error_log(
                "The input of type argument does not support, only supports tf, caffe, quant and offline.")
            return CompareError.MSACCUCMP_INVALID_TYPE_ERROR
        # remove old error file list
        error_file_path = os.path.join(self.output_path, ConstManager.CONVERT_FAILED_FILE_LIST_NAME)
        if os.path.exists(error_file_path):
            os.remove(error_file_path)
        return return_code

    def convert_data(self: any) -> int:
        """
        Convert data between numpy and dump
        :return: error_code
        """
        # 1. check arguments valid
        return_code = self.check_arguments_valid()
        if return_code != CompareError.MSACCUCMP_NONE_ERROR:
            return return_code
        # 2. convert dump data one by one
        if os.path.isfile(self.input_path):
            ret, _ = self._convert_file(self.input_path)
            return ret
        return self.multi_process.process()

    def _get_op_name_from_path(self: any, file_path: str) -> str:
        layer_name = os.path.basename(file_path)
        layer_name = self.layer_name_switch[self.type](layer_name) if \
            self.type in self.layer_name_switch else layer_name
        return layer_name

    def _get_standard_layer_name(self: any, layer_name: str) -> str:
        match = None
        if self.target == "numpy":
            _, match = RegManager.match_group(RegManager.STANDARD_DUMP_PATTERN, layer_name)
            if match is None:
                log.print_warn_log(
                    "Invalid file name: %s, only supports 'op_name.output_index.timestamp.pb'." % layer_name)
                raise CompareError(CompareError.MSACCUCMP_DUMP_FILE_ERROR)
        elif self.target == "dump":
            _, match = RegManager.match_group(RegManager.STANDARD_NUMPY_PATTERN, layer_name)
            if match is None:
                log.print_warn_log(
                    "Invalid file name: %s, only supports 'op_name.output_index.timestamp.npy'." % layer_name)
                raise CompareError(CompareError.MSACCUCMP_DUMP_FILE_ERROR)
        return match.group(1)

    def _get_quant_layer_name(self: any, layer_name: str) -> str:
        match = None
        if self.target == "numpy":
            _, match = RegManager.match_group(RegManager.QUANT_DUMP_PATTERN, layer_name)
            if match is None:
                log.print_warn_log(
                    "Invalid file name: %s, only supports 'op_name.output_index.timestamp.quant'." % layer_name)
                raise CompareError(CompareError.MSACCUCMP_DUMP_FILE_ERROR)
        elif self.target == "dump":
            _, match = RegManager.match_group(RegManager.STANDARD_NUMPY_PATTERN, layer_name)
            if match is None:
                log.print_warn_log(
                    "Invalid file name: %s, only supports 'op_name.output_index.timestamp.npy'." % layer_name)
                raise CompareError(CompareError.MSACCUCMP_DUMP_FILE_ERROR)
        return match.group(1)

    def _get_offline_layer_name(self: any, layer_name: str) -> str:
        match = None
        if self.target == "numpy":
            _, match = RegManager.match_group(RegManager.OFFLINE_DUMP_PATTERN, layer_name)
            if match is None:
                log.print_warn_log(
                    "Invalid file name: %s, only supports 'op_type.op_name.task_id.timestamp'." % layer_name)
                raise CompareError(CompareError.MSACCUCMP_DUMP_FILE_ERROR)
        elif self.target == "dump":
            _, match = RegManager.match_group(RegManager.OFFLINE_NUMPY_PATTERN, layer_name)
            if match is None:
                log.print_warn_log(
                    "Invalid file name: %s, only supports 'op_type.op_name.task_id.timestamp."
                    "output_index.npy/data/bin/txt'." % layer_name)
                raise CompareError(CompareError.MSACCUCMP_DUMP_FILE_ERROR)
        return match.group(1)

    def _save_tensor_to_file(self: any, tensor_list: list, name: str, input_path: str, tensor_type: str) -> None:
        if len(tensor_list) == 0:
            log.print_warn_log('There is no %s in "%r".' % (tensor_type, input_path))
            return
        for (index, tensor) in enumerate(tensor_list):
            log.print_info_log('Start to convert %s:%d of "%r" to numpy file.'
                               % (tensor_type, index, input_path))
            array = tensor.data
            try:
                array = array.reshape(tensor.shape)
            except ValueError as error:
                log.print_error_log(
                    'Failed to convert %s:%d of "%r" to numpy file. %s' % (tensor_type, index, input_path, error))
                continue
            finally:
                pass

            if self.type == "caffe" or self.type == "tf" or self.type == "quant":
                file_name = "%s.npy" % name
            else:
                file_name = "%s.%s.%d.npy" % (name, tensor_type, index)

            output_dump_path = os.path.join(self.output_path, file_name)
            np.save(output_dump_path, array)
            log.print_info_log('The %s:%d of "%r" has been converted to file "%r".'
                               % (tensor_type, index, input_path, output_dump_path))

    def _save_buffer_to_file(self: any, tensor_list: list, name: str, input_path: str) -> None:
        if len(tensor_list) == 0:
            log.print_warn_log('There is no buffer data in "%r".' % input_path)
            return
        for (index, tensor) in enumerate(tensor_list):
            buffer_type = ConstManager.BUFFER_TYPE_MAP.get(tensor.buffer_type)
            log.print_info_log('Start to convert %sbuffer:%d of "%r" to bin file.'
                               % (buffer_type, index, input_path))
            file_name = "%s.%sbuffer.%s.bin" % (name, buffer_type, index)
            output_dump_path = os.path.join(self.output_path, file_name)
            try:
                with os.fdopen(os.open(output_dump_path, ConstManager.WRITE_FLAGS,
                                       ConstManager.WRITE_MODES), 'wb') as output_file:
                    output_file.write(tensor.data)
                    log.print_info_log('The %sbuffer:%d of "%r" has been converted to file "%r".'
                                       % (buffer_type, index, input_path, output_dump_path))
            except IOError as io_error:
                log.print_error_log('Failed to open "%r". %s ' % (output_dump_path, str(io_error)))
            finally:
                pass

    def _save_space_to_file(self: any, tensor_list: list, name: str, input_path: str) -> None:
        if len(tensor_list) == 0:
            log.print_warn_log('There is no space data in "%r".' % input_path)
            return
        for (index, tensor) in enumerate(tensor_list):
            file_name = "%s.space.%s.bin" % (name, index)
            log.print_info_log('Start to save space:%d of "%r" to bin file.' % (index, input_path))
            output_dump_path = os.path.join(self.output_path, file_name)
            try:
                with os.fdopen(os.open(output_dump_path, ConstManager.WRITE_FLAGS,
                               ConstManager.WRITE_MODES), 'wb') as output_file:
                    output_file.write(tensor.data)
                    log.print_info_log('The space:%d of "%r" has been converted to file "%r".'
                               % (index, input_path, output_dump_path))
            except IOError as io_error:
                log.print_error_log('Failed to open "%r". %s ' % (output_dump_path, str(io_error)))
            finally:
                pass

    def _convert_file_exec(self: any, input_path: str) -> None:
        name = self._get_op_name_from_path(input_path)
        if self.target == "dump":
            log.print_info_log('Start to convert the numpy file "%r" to dump file.' % input_path)
            numpy_data = dump_utils.read_numpy_file(input_path)
            file_name = self.file_name_switch[self.type](name) if self.type in self.file_name_switch else ""

            output_dump_path = os.path.join(self.output_path, file_name)
            big_dump_data.write_dump_data(numpy_data, output_dump_path)
            log.print_info_log(
                'The file "%r" has been converted to file "%r".' % (input_path, output_dump_path))
        elif self.target == "numpy":
            dump_data = dump_utils.parse_dump_file(input_path, ConstManager.OLD_DUMP_TYPE)
            self._save_tensor_to_file(dump_data.input_data, name, input_path, 'input')
            self._save_tensor_to_file(dump_data.output_data, name, input_path, 'output')
            self._save_buffer_to_file(dump_data.buffer, name, input_path)
            self._save_space_to_file(dump_data.space, name, input_path)

    def _convert_file(self: any, input_path: str) -> (int, str):
        return_code = CompareError.MSACCUCMP_NONE_ERROR
        try:
            self._convert_file_exec(input_path)
        except CompareError as error:
            return_code = error.code
        except MemoryError:
            log.print_error_log('Failed to convert file "%r" by memory error.' % input_path)
            return_code = CompareError.MSACCUCMP_UNKNOWN_ERROR
        finally:
            pass
        return return_code, input_path



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import os
import sys
import math
import argparse
import ast
import heapq
from abc import ABC
from dataclasses import dataclass, field
from typing import List

from msprobe.core.common.const import MonitorConst
from msprobe.core.common.log import logger
from msprobe.core.common.file_utils import save_json, create_directory, remove_path, \
    check_file_or_directory_path, load_json


class ScanRule(ABC):
    name = "ScanRule"

    def apply(self, cur, history=None):
        raise NotImplementedError("abstract method apply is not implemented")


class AnomalyTurbulence(ScanRule):
    name = "AnomalyTurbulence"

    def __init__(self, threshold) -> None:
        self.threshold = threshold

    def apply(self, cur, history=None):
        """
        :param cur: float, current metric value
        :param history: float, history weighted average
        :return: bool, whether the current value deviates from the historical average value of current metric
        """
        up_bound = history * (1 + self.threshold)
        return abs(cur) > up_bound


class AnomalyNan(ScanRule):
    name = "AnomalyNan"

    def __init__(self, threshold=None) -> None:
        self.threshold = threshold

    def apply(self, cur, history=None):
        return math.isnan(cur) or (self.threshold is not None and abs(cur) > self.threshold)


class AnomalyScanner:

    @staticmethod
    def load_rules(specs: List[dict]):
        """
        specs: [{"rule_name": "AnomalyTurbulence", "args": {"threshold": 0.5}}]
        """
        if specs is None:
            return []
        alert_rules = []
        for spec in specs:
            # 使用get方法获取键值，如果键不存在则返回None
            rule_cls_name = spec.get("rule_name")
            rule_args = spec.get("args")

            # 检查必要的键是否存在
            if rule_cls_name is None or (rule_cls_name == "AnomalyTurbulence" and rule_args is None):
                logger.warning(f"Spec is missing required keys: {spec}")
                continue

            cur_module = sys.modules.get(__name__)
            try:
                rule_cls = getattr(cur_module, rule_cls_name)
            except AttributeError:
                logger.error(f"Rule class '{rule_cls_name}' not found in the current module.")
                continue

            try:
                rule_instance = rule_cls(**rule_args) if rule_args is not None else rule_cls()
                alert_rules.append(rule_instance)
            except Exception as e:
                logger.error(f"Error creating instance of rule '{rule_cls_name}': {e}")
                continue

        return alert_rules

    @staticmethod
    def scan(scan_rules: List[ScanRule], history, cur):
        anomaly = False
        for rule in scan_rules:
            anomaly = rule.apply(cur, history=history)
            if anomaly:
                return anomaly, rule.name
        return anomaly, None


class AnomalyDataFactory(ABC):
    def __init__(self, rank, pp_stage, group_mates):
        super().__init__()
        self.rank = rank
        self.pp_stage = pp_stage
        self.group_mates = group_mates
        self.micro_step = 0
        self.name2callid = {}

    def set_call_id(self, name2callid):
        """根据当前GradContext信息更新call_id vpp_stage等信息
        """
        self.name2callid = name2callid

    def create(self, tag, message, step):
        """如果检查出异常, 调用当前接口生成GradAnomalyData实例
        tag (tuple): metric tag ('0:1.post_attention_norm.weight/rank0/pre_grad', 'min')
        message (str): anomaly detect message
        step (int): training step
        """
        if not isinstance(tag, tuple) or len(tag) != 2:
            raise ValueError("tag must be a tuple with length 2")
        tag_name = tag[0]
        param_name = tag_name.split('/')[0]
        call_id = self.name2callid.get(tag_name, -1)
        if MonitorConst.NAME_SEP in param_name:
            vpp_stage = int(param_name.split(MonitorConst.NAME_SEP)[0])
        else:
            vpp_stage = 0

        return GradAnomalyData(
            self.rank,
            step,
            self.micro_step,
            self.pp_stage,
            vpp_stage,
            call_id,
            tag_name,
            message,
            self.group_mates
        )


@dataclass(eq=True)
class GradAnomalyData:
    rank: int = 0
    step: int = 0
    micro_step: int = 0
    pp_stage: int = 0
    vpp_stage: int = 0
    call_id: int = 0
    tag_name: str = field(default=None, compare=False)
    message: str = field(default="", compare=False)
    group_mates: list = field(default=None, compare=False)

    def __lt__(self, other):
        """
        自定义比较函数，用于确定 GradAnomalyData 实例之间的顺序。
        比较规则为：
            step 和 micro_step 值越小优先级越高；
            vpp 和 pp 在前向阶段值越小优先级越高，在非前向阶段值越大优先级越高；
            call_id 值越小优先级越高。
        """
        if not isinstance(other, GradAnomalyData):
            return NotImplemented

        self_train_stage = self.get_train_stage(self.tag_name)
        other_train_stage = self.get_train_stage(other.tag_name)

        def vpp_pp_comparator(anomaly):
            """
            Determine the priority rule for vpp and pp based on train stage
            Forward stage prefers smaller vpp and pp
            Other stages prefer larger vpp and pp
            """
            if self_train_stage == MonitorConst.FORWARD_STAGE:
                return anomaly.vpp_stage, anomaly.pp_stage
            else:
                return -anomaly.vpp_stage, -anomaly.pp_stage

        self_cmp = [self.step, self.micro_step, self_train_stage, *vpp_pp_comparator(self), self.call_id]
        other_cmp = [other.step, other.micro_step, other_train_stage, *vpp_pp_comparator(other), other.call_id]
        return self_cmp < other_cmp

    def __le__(self, other):
        if not isinstance(other, GradAnomalyData):
            return NotImplemented
        return self == other or self < other

    @staticmethod
    def get_train_stage(tag_name):
        """
        :param tag_name: "0:fc2.input:0/rank0/actv", "0:fc1.weight/rank0/post_grad", "0:fc2.weight/rank0/exp_avg_sq"
        :return: int, if forward return 0; if backward return 1; if optimizer return 2
        """
        key_ = tag_name.split("/")[-1]
        return MonitorConst.TRAIN_STAGE.get(key_, MonitorConst.DEFAULT_STAGE)

    def to_dict(self):
        return self.__dict__

    def get_key(self):
        # 0:1.self_attention.core_attention_flash_0/rank0/input_grad
        return ''.join([str(self.tag_name), "_step_", str(self.step), "_call_", str(self.call_id)])


class AnomalyDataWriter:
    """
    异常数据写入类，负责将异常数据写入到JSON文件中。
    """

    def __init__(self, dump_path, rank) -> None:
        self.dump_path = dump_path
        self.dump_rank_dir = os.path.join(self.dump_path, f"rank{rank}")
        self.json_path = os.path.join(self.dump_rank_dir, MonitorConst.ANOMALY_JSON)

    @staticmethod
    def get_anomaly_dict(anomalies):
        """将GradAnomalyData列表转换为json"""
        anomalies_json = {}
        for anomaly in anomalies:
            anomalies_json.update({anomaly.get_key(): anomaly.to_dict()})
        return anomalies_json

    def init_detected_json(self):
        """初始化落盘文件"""
        create_directory(self.dump_rank_dir)

        if os.path.exists(self.json_path):
            check_file_or_directory_path(self.json_path, isdir=False)
            logger.warning(f"The existing file will be deleted: {self.json_path}.")
            remove_path(self.json_path)
        save_json(self.json_path, {}, indent=1)

    def write_detected_json(self, anomalies):
        """
        落盘异常数据
        Args:
        anomalies: GradAnomalyData对象列表
        """
        anomalies_json = self.get_anomaly_dict(anomalies)
        if anomalies_json:
            logger.info(f"{MonitorConst.ANOMALY_JSON} is at {self.dump_rank_dir}.")

            data_to_write = load_json(self.json_path) if os.path.exists(self.json_path) else {}
            data_to_write.update(anomalies_json)
            save_json(self.json_path, data_to_write, indent=1)


class AnomalyDataLoader:
    def __init__(self, data_path) -> None:
        self.data_path = data_path

    @staticmethod
    def create_instances_from_dict(anomalies_dict: dict):
        instances = []
        for values in anomalies_dict.values():
            try:
                instances.append(GradAnomalyData(**values))
            except KeyError as e:
                logger.warning(f"Missing key in anomaly data: {e}.")
            except Exception as e:
                logger.warning(f"Value error when creating a GradAnomalyData instance: {e}.")
        return instances

    def get_anomalies_from_jsons(self):
        """遍历文件夹,从rankK/anomaly.json中读取异常数据
        return: anomalies: GradAnomalyData对象列表
        """
        anomalies = []
        check_file_or_directory_path(self.data_path, isdir=True)
        for rank_dir in os.listdir(self.data_path):
            rank_path = os.path.join(self.data_path, rank_dir)
            if not os.path.isdir(rank_path):
                continue
            json_path = os.path.join(rank_path, MonitorConst.ANOMALY_JSON)
            if not os.path.exists(json_path):
                continue
            data_anomalies = load_json(json_path)
            instances = self.create_instances_from_dict(data_anomalies)
            anomalies.extend(instances)
        return anomalies


class AnomalyAnalyse:
    def __init__(self) -> None:
        self.sorted_anomalies = []

    def get_range_top_k(self, topk, step_list, anomalies):
        """
        获取前topk个step_list范围内的异常。
        """
        if not step_list:
            filtered_anomalies = anomalies
        else:
            filtered_anomalies = [
                anomaly
                for anomaly in anomalies
                if anomaly.step in step_list
            ]
        if topk >= len(filtered_anomalies):
            self.sorted_anomalies = sorted(filtered_anomalies)
        else:
            self.sorted_anomalies = list(heapq.nsmallest(topk, filtered_anomalies))
        return self.sorted_anomalies

    def rewrite_sorted_anomalies(self, output_path):
        """
        将排序后的异常数据重新落盘
        """
        check_file_or_directory_path(output_path, isdir=True)

        sorted_data = AnomalyDataWriter.get_anomaly_dict(self.sorted_anomalies)
        logger.info(f"{MonitorConst.ANALYSE_JSON} is at {output_path}.")
        json_path = os.path.join(output_path, MonitorConst.ANALYSE_JSON)
        if os.path.exists(json_path):
            logger.warning(f"The existing file will be deleted: {json_path}.")
            remove_path(json_path)
        save_json(json_path, sorted_data, indent=1)


def _get_step_and_stop(args):
    try:
        step_list = ast.literal_eval(args.step_list)
        if not isinstance(step_list, list):
            raise ValueError(f"{args.step_list} is not a list.")
    except (ValueError, SyntaxError, RecursionError) as e:
        raise Exception(f"The step list must be a resolvable list type.") from e
    if args.top_k_number <= 0:
        raise Exception("The top k number must be greater than 0.")
    return step_list, args.top_k_number


def _anomaly_analyse():
    args = _get_parse_args()
    step_list, top_k_number = _get_step_and_stop(args)
    loader = AnomalyDataLoader(args.data_path_dir)
    anomalies = loader.get_anomalies_from_jsons()
    analyser = AnomalyAnalyse()
    top_anomalies = analyser.get_range_top_k(
        top_k_number, step_list, anomalies
    )
    analyser.rewrite_sorted_anomalies(
        args.out_path if args.out_path else args.data_path_dir
    )

    logger.info(f"Top {top_k_number} anomalies are listed as follows:")
    for index, anomaly in enumerate(top_anomalies):
        logger.info(f"{index}: {anomaly.message}")


def _get_parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("-d", "--data_path", dest="data_path_dir", default="./", type=str,
                        help="<Required> The anomaly detect result dictionary: generate from monitor tool.",
                        required=True,
                        )
    parser.add_argument("-o", "--out_path", dest="out_path", default="", type=str,
                        help="<optional> The analyse task result out path.",
                        required=False,
                        )
    parser.add_argument("-k", "--topk", dest="top_k_number", default=8, type=int,
                        help="<optional> Top K number of earliest anomalies.",
                        required=False,
                        )
    parser.add_argument("-s", "--step", dest="step_list", default="[]", type=str,
                        help="<optional> Analyse which steps.",
                        required=False,
                        )
    return parser.parse_args(sys.argv[1:])


if __name__ == "__main__":
    _anomaly_analyse()
    logger.info("Analyse task completed.")



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import os
import types
from collections import defaultdict

import mindspore
from mindspore import nn
from mindspore._c_expression import PyNativeExecutor_

try:
    from mindspore.common.api import _MindsporeFunctionExecutor
except ImportError:
    from mindspore.common.api import _JitExecutor as _MindsporeFunctionExecutor

from msprobe.core.common.log import logger
from msprobe.core.common.const import Const
from msprobe.core.common.utils import ThreadSafe
from msprobe.core.common.runtime import Runtime
from msprobe.core.dump.data_dump.data_processor.base import ModuleForwardInputsOutputs, ModuleBackwardInputsOutputs
from msprobe.mindspore.common.const import Const as MsConst
from msprobe.mindspore.dump.dump_processor.hook_cell.api_register import get_api_register

_api_register = get_api_register()


def dump_jit(name, in_feat, out_feat, is_forward):
    pid = os.getpid()
    name = name if name else "JitFunction"
    if not JitDump.need_dump():
        return
    with ThreadSafe():
        if is_forward:
            if name in JitDump.jit_count:
                JitDump.jit_count[name] += 1
            else:
                JitDump.jit_count[name] = 0
            name_template = (Const.JIT + Const.SEP + name + Const.SEP +
                             str(JitDump.jit_count[name]) + Const.SEP + Const.FORWARD)
            JitDump.data_collector.update_api_or_module_name(name_template)
            module_input_output = ModuleForwardInputsOutputs(args=in_feat, kwargs={}, output=out_feat)
            JitDump.data_collector.forward_data_collect(name_template, None, pid, module_input_output)
        else:
            name_template = Const.JIT + Const.SEP + name + Const.SEP + str(JitDump.jit_count[name]) + Const.SEP + \
                            Const.BACKWARD
            JitDump.data_collector.update_api_or_module_name(name_template)
            module_input_output = ModuleBackwardInputsOutputs(grad_input=in_feat, grad_output=out_feat)
            JitDump.data_collector.backward_data_collect(name_template, None, pid, module_input_output)


class JitDump(_MindsporeFunctionExecutor):
    dump_config = None
    jit_enable = False
    jit_dump_switch = False
    jit_count = defaultdict(int)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = None
        if len(args) > 0:
            self.name = args[0].__name__
        self._executor = PyNativeExecutor_.get_instance()

    def __call__(self, *args, **kwargs):
        _api_register.restore_all_api()
        out = super().__call__(*args, **kwargs)
        if JitDump.jit_dump_switch and len(args) > 0 and self.name:
            if self.name != "construct":
                dump_jit(self.name, args, out, True)
            elif Runtime.run_mode != MsConst.PYNATIVE_GRAPH_MODE and isinstance(args[0], nn.Cell):
                dump_jit(args[0].__class__.__name__, args, out, True)
            JitDump.jit_enable = True
        elif len(args) == 0:
            logger.warning(f"The jit function {self.name} has no input arguments, nothing will be dumped.")
        _api_register.register_all_api()
        return out

    @classmethod
    def set_config(cls, value):
        cls.dump_config = value

    @classmethod
    def set_data_collector(cls, value):
        cls.data_collector = value

    @classmethod
    def need_dump(cls):
        if cls.dump_config.task != Const.TENSOR and cls.dump_config.task != Const.STATISTICS:
            return False
        if not cls.data_collector or cls.data_collector.data_processor.is_terminated:
            return False
        return True

    def grad(self, obj, grad, weights, grad_position, *args, **kwargs):
        if JitDump.jit_dump_switch and JitDump.jit_enable:
            _api_register.restore_all_api()
        if mindspore.__version__ >= "2.5":
            output = self._executor.grad(grad, obj, weights, grad_position, False, *args, *(kwargs.values()))
        else:
            output = self._executor.grad(grad, obj, weights, grad_position, *args, *(kwargs.values()))
        if JitDump.jit_dump_switch and JitDump.jit_enable:
            if isinstance(obj, types.FunctionType):
                dump_jit(obj.__name__, args, None, False)
            elif Runtime.run_mode != MsConst.PYNATIVE_GRAPH_MODE and isinstance(obj, nn.Cell):
                dump_jit(obj.__class__.__name__, args, None, False)
            _api_register.register_all_api()
        return output



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import os
import sys
import stat
import re
import logging
from enum import Enum

from msprobe.core.common.log import logger
from msprobe.infer.utils.constants import PATH_WHITE_LIST_REGEX
from msprobe.infer.utils.check import Rule
from msprobe.infer.utils.constants import CONFIG_FILE_MAX_SIZE

MAX_SIZE_UNLIMITE = -1  # 不限制，必须显式表示不限制，读取必须传入
MAX_SIZE_LIMITE_CONFIG_FILE = 10 * 1024 * 1024  # 10M 普通配置文件，可以根据实际要求变更
MAX_SIZE_LIMITE_NORMAL_FILE = 4 * 1024 * 1024 * 1024  # 4G 普通模型文件，可以根据实际要求变更
MAX_SIZE_LIMITE_MODEL_FILE = 100 * 1024 * 1024 * 1024  # 100G 超大模型文件，需要确定能处理大文件，可以根据实际要求变更

PATH_WHITE_LIST_REGEX_WIN = re.compile(r"[^_:\\A-Za-z0-9/.-]")

PERMISSION_NORMAL = 0o640  # 普通文件
PERMISSION_KEY = 0o600  # 密钥文件
READ_FILE_NOT_PERMITTED_STAT = stat.S_IWGRP | stat.S_IWOTH
WRITE_FILE_NOT_PERMITTED_STAT = stat.S_IWGRP | stat.S_IWOTH

SOLUTION_LEVEL = 35
SOLUTION_LEVEL_WIN = 45
logging.addLevelName(SOLUTION_LEVEL, "\033[1;32m" + "SOLUTION" + "\033[0m")  # green [SOLUTION]
logging.addLevelName(SOLUTION_LEVEL_WIN, "SOLUTION_WIN")

RAW_INPUT_PATH = "RAW_INPUT_PATH"

MALICIOUS_CSV_PATTERN = re.compile(r'^[＝＋－+-=%@];[＝＋－+-=%@]')


def is_legal_path_length(path):
    if len(path) > 4096 and not sys.platform.startswith("win"):  # linux total path length limit
        logger.error(f"file total path {path} length out of range (4096), please check the file(or directory) path")
        return False

    if len(path) > 260 and sys.platform.startswith("win"):  # windows total path length limit
        logger.error(f"file total path {path} length out of range (260), please check the file(or directory) path")
        return False

    dirnames = path.split("/")
    for dirname in dirnames:
        if len(dirname) > 255:  # linux single file path length limit
            logger.error(f"file name {dirname} length out of range (255), please check the file(or directory) path")
            return False
    return True


def is_match_path_white_list(path):
    if PATH_WHITE_LIST_REGEX.search(path) and not sys.platform.startswith("win"):
        logger.error(f"path: {path} contains illegal char, legal chars include A-Z a-z 0-9 _ - / .")
        return False
    if PATH_WHITE_LIST_REGEX_WIN.search(path) and sys.platform.startswith("win"):
        logger.error(f"path: {path} contains illegal char, legal chars include A-Z a-z 0-9 _ - / . : \\")
        return False
    return True


def is_legal_args_path_string(path):
    # only check path string
    if not path:
        return True
    if not is_legal_path_length(path):
        return False
    if not is_match_path_white_list(path):
        return False
    return True


class SanitizeErrorType(Enum):
    """
    The errors parameter Enum of the function sanitize_csv_value
    """
    strict = "strict"
    ignore = "ignore"
    replace = "replace"


def sanitize_csv_value(value: str, errors=SanitizeErrorType.strict.value):
    if errors == SanitizeErrorType.ignore.value or not isinstance(value, str):
        return value

    sanitized_value = value
    try:
        float(value)  # in case value is a digit but in str format
    except ValueError as e:  # not digit
        if not MALICIOUS_CSV_PATTERN.search(value):
            pass
        elif errors == SanitizeErrorType.replace.value:
            sanitized_value = ' ' + value
        else:
            msg = f'Malicious value is not allowed to be written to the csv {value}'
            logger.error("Please check the value written to the csv")
            raise ValueError(msg) from e

    return sanitized_value


class OpenException(Exception):
    pass


class FileStat:
    def __init__(self, file) -> None:
        if not is_legal_path_length(file) or not is_match_path_white_list(file):
            raise OpenException("Path name is too long or contains invalid characters.")
        self.file = file
        self.is_file_exist = os.path.exists(file)
        if self.is_file_exist:
            self.file_stat = os.stat(file)
            self.realpath = os.path.realpath(file)
        else:
            self.file_stat = None

    @property
    def is_exists(self):
        return self.is_file_exist

    @property
    def is_softlink(self):
        return os.path.islink(self.file) if self.file_stat else False

    @property
    def is_file(self):
        return stat.S_ISREG(self.file_stat.st_mode) if self.file_stat else False

    @property
    def is_dir(self):
        return stat.S_ISDIR(self.file_stat.st_mode) if self.file_stat else False

    @property
    def file_size(self):
        return self.file_stat.st_size if self.file_stat else 0

    @property
    def permission(self):
        return stat.S_IMODE(self.file_stat.st_mode) if self.file_stat else 0o777

    @property
    def owner(self):
        return self.file_stat.st_uid if self.file_stat else -1

    @property
    def group_owner(self):
        return self.file_stat.st_gid if self.file_stat else -1

    @property
    def is_owner(self):
        return self.owner == (os.geteuid() if hasattr(os, "geteuid") else 0)

    @property
    def is_group_owner(self):
        return self.group_owner in (os.getgroups() if hasattr(os, "getgroups") else [0])

    @property
    def is_user_or_group_owner(self):
        return self.is_owner or self.is_group_owner

    @property
    def is_user_and_group_owner(self):
        return self.is_owner and self.is_group_owner

    def check_owner_or_root(self):
        if os.getuid() == self.file_stat.st_uid:
            return True
        elif os.getuid() == 0:
            logger.warning("You are currently operating this tool using the root user. "
                           "Please be aware of the risk of privilege escalation.")
            return True
        else:
            logging.error("The file owner is not consistent with the current user.")
            return False

    def is_basically_legal(self, perm='none', strict_permission=True):
        if sys.platform.startswith("win"):
            return self.check_windows_permission(perm)
        else:
            return self.check_linux_permission(perm, strict_permission=strict_permission)

    def check_basic_permission(self, perm='none'):
        if not self.is_exists and perm != 'write':
            logger.error(f"path: {self.file} not exist, please check if file or dir is exist")
            return False
        if self.is_softlink:
            whitelist_path = os.environ.get(RAW_INPUT_PATH, "")
            if whitelist_path == "":
                logger.error(f"path : {self.file} is a soft link, not supported, "
                             f"please import file(or directory) directly")
                return False
            target = os.readlink(self.file)
            target_path = os.path.abspath(os.path.normpath(target))  # normpath更加规范
            file_path = os.path.abspath(os.path.normpath(self.file))
            sub_paths = whitelist_path.split("|")
            illegal_softlink = True
            for sub_path in sub_paths:
                sub_path_abs = os.path.abspath(os.path.normpath(sub_path))
                # 检查子路径本身是否是软链接
                if os.path.islink(sub_path_abs):
                    continue
                # 使用 os.path.commonpath 来比较路径
                common_path_target = os.path.commonpath([sub_path_abs, target_path])
                common_path_file = os.path.commonpath([sub_path_abs, file_path])
                # 确保公共路径与子路径相同，表示目标路径和文件路径都在子路径内
                if common_path_target == sub_path_abs and common_path_file == sub_path_abs:
                    illegal_softlink = False
                    break  # 已找到合法路径，退出循环
            if illegal_softlink:
                logger.error(f"path : {self.file} is a soft link, not supported, "
                             f"please import file(or directory) directly")
                return False
        return True

    def check_linux_permission(self, perm='none', strict_permission=True):
        if not self.check_basic_permission(perm=perm):
            return False
        if not self.is_user_or_group_owner and self.is_exists:
            logger.error(f"current user isn't path: {self.file}'s owner or ownergroup")
            return False
        if self.is_exists and not self.check_owner_or_root():
            return False
        if perm == 'read':
            if strict_permission and self.permission & READ_FILE_NOT_PERMITTED_STAT > 0:
                logger.error(f"The file {self.file} is group writable, or is others writable, "
                             "as import file(or directory) permission should not be over 0o755(rwxr-xr-x)")
                return False
            if not os.access(self.realpath, os.R_OK) or self.permission & stat.S_IRUSR == 0:
                logger.error(f"Current user doesn't have read permission to the file {self.file}, "
                             "as import file(or directory) permission should be at least 0o400(r--------)")
                return False
        elif perm == 'write' and self.is_exists:
            if (strict_permission or self.is_file) and self.permission & WRITE_FILE_NOT_PERMITTED_STAT > 0:
                logger.error(f"The file {self.file} is group writable, or is others writable, "
                             "as export file(or directory) permission should not be over 0o755(rwxr-xr-x)")
                return False
            if not os.access(self.realpath, os.W_OK):
                logger.error(f"Current user doesn't have write permission to the file {self.file}, "
                             "as export file(or directory) permission should be at least 0o200(-w-------)")
                return False
        return True

    def check_windows_permission(self, perm='none'):
        if not self.check_basic_permission(perm=perm):
            return False
        return True

    def is_legal_file_size(self, max_size):
        if not self.is_file:
            logger.error(f"path: {self.file} is not a file")
            return False
        if self.file_size > max_size:
            logger.error(f"file_size: {self.file_size} byte out of max limit {max_size} byte")
            return False
        else:
            return True

    def is_legal_file_type(self, file_types: list):
        if not self.is_file and self.is_exists:
            logger.error(f"path: {self.file} is not a file")
            return False
        for file_type in file_types:
            if os.path.splitext(self.file)[1] == f".{file_type}":
                return True
        logger.error(f"path: {self.file}, file type not in {file_types}")
        return False


def ms_open(file, mode="r", max_size=CONFIG_FILE_MAX_SIZE, softlink=False,
            write_permission=PERMISSION_NORMAL, **kwargs):
    file_stat = FileStat(file)

    if file_stat.is_exists and file_stat.is_dir:
        raise OpenException(f"Expecting a file, but it's a folder. {file}")

    if file_stat.is_exists and not file_stat.check_owner_or_root():
        raise OpenException(f"There is a problem with the owner of the file. Please check it.")

    if "r" in mode:
        if not file_stat.is_exists:
            raise OpenException(f"No such file or directory {file}")
        if max_size is None:
            raise OpenException(f"Reading files must have a size limit control. {file}")
        if max_size != MAX_SIZE_UNLIMITE and max_size < file_stat.file_size:
            raise OpenException(f"The file size has exceeded the specifications and cannot be read. {file}")

    if "w" in mode and file_stat.is_exists:
        if not file_stat.is_owner:
            raise OpenException(
                f"The file owner is inconsistent with the current process user and is not allowed to write. {file}"
            )
        os.remove(file)

    if not softlink and file_stat.is_softlink:
        raise OpenException(f"Softlink is not allowed to be opened. {file}")

    if "a" in mode and file_stat.is_exists:
        if not file_stat.is_owner:
            raise OpenException(
                f"The file owner is inconsistent with the current process user and is not allowed to write. {file}"
            )
        if file_stat.permission != (file_stat.permission & write_permission):
            os.chmod(file, file_stat.permission & write_permission)

    safe_parent_msg = Rule.path().is_safe_parent_dir().check(file)
    if not safe_parent_msg:
        logger.warning(f"parent dir of {os.path.realpath(file)} is not safe. {str(safe_parent_msg)}")

    if "+" in mode:
        flags = os.O_RDONLY | os.O_RDWR
    elif "w" in mode or "a" in mode or "x" in mode:
        flags = os.O_RDONLY | os.O_WRONLY
    else:
        flags = os.O_RDONLY

    if "w" in mode or "x" in mode:
        flags = flags | os.O_TRUNC | os.O_CREAT
    if "a" in mode:
        flags = flags | os.O_APPEND | os.O_CREAT
    return os.fdopen(os.open(file, flags, mode=write_permission), mode, **kwargs)



# coding=utf-8
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
Function:
This file mainly involves the reg const.
"""

import re

from msprobe.msaccucmp.cmp_utils.constant.const_manager import ConstManager


class RegManager:
    """
    The class for reg manager
    """
    NUMBER_PATTERN = r"^[0-9]+$"

    # mapping of built in algorithms to numbers, the indexes correspond one-to-one to the list above
    BUILTIN_ALGORITHM_INDEX_PATTERN = r"^([0-" + str(len(ConstManager.BUILT_IN_ALGORITHM) - 1) + "])$"

    # Standard
    STANDARD_DUMP_PATTERN = r"^([A-Za-z0-9_-]+\.[0-9]+)\.[0-9]{1,255}\.pb$"

    # Qunat
    QUANT_DUMP_PATTERN = r"^([A-Za-z0-9_-]+\.[0-9]+)\.[0-9]{1,255}\.quant$"

    # Offline
    OFFLINE_DUMP_PATTERN = r"^[A-Za-z0-9_-]+\.([A-Za-z0-9_-]+)\.[0-9]+" \
                           r"(\.[0-9]+)?\.[0-9]{1,255}(\.[0-9]+\.[0-9]+\.[0-9]+)?(\.[0-9]+)?"
    OFFLINE_NUMPY_PATTERN = r"^([A-Za-z0-9_-]+\.[A-Za-z0-9_-]+\.[0-9]+" \
                            r"\.[0-9]{1,255})\.[0-9]+\.\b(npy|data|bin|txt)\b$"
    OFFLINE_FFTS_DUMP_PATTERN = r"^[A-Za-z0-9_-]+\.([A-Za-z0-9_-]+)\.[0-9]+" \
                                r"(\.[0-9]+)?\.[0-9]{1,255}\.[0-9]+\.[0-9]+\.[0-9]+"

    # Standard
    NUMPY_DUMP_PATTERN = r"^([\.A-Za-z0-9_-]+\.[0-9]+)\.[0-9]{1,255}\.npy$"

    STANDARD_NUMPY_PATTERN = r"^([\.A-Za-z0-9_-]+\.[0-9]+\.[0-9]{1,255})\.npy$"

    SUPPORT_SHAPE_PATTERN = r"^([0-9]+,)+[0-9]+$"

    # matching `convert_xxx_to_yyy.py`, the `{1,128}` is used for limiting string length, bypassing ReDoS attack.
    FORMAT_CONVERT_FILE_NAME_PATTERN = r"^(convert_[A-Za-z0-9_]{1,128}_to_[A-Za-z0-9_]{1,128})\.py[c]?$"

    SUPPORT_PATH_PATTERN = r"^[A-Za-z0-9_\./:()=\\-]+$"

    FFTS_MANUAL_FIELD_PATTERN = r"lxslice[0-9]+"

    LXSLICE_PATTERN = r"_lxslice[0-9]+"

    SGT_FLIED_PATTERN = r"sgt_graph_[0-9]+"

    @staticmethod
    def match_pattern(pattern: str, value: any) -> bool:
        """
        The value match pattern or not
        :param pattern: the pattern
        :param value: the value to match
        :return bool
        """
        re_pattern = re.compile(pattern)
        match = re_pattern.match(value)
        return match is not None

    @staticmethod
    def match_group(pattern: str, value: any) -> (bool, any):
        """
        The value match pattern or not
        :param pattern: the pattern
        :param value: the value to match
        :return bool, match
        """
        re_pattern = re.compile(pattern)
        match = re_pattern.match(value)
        if match is not None:
            return True, match
        return False, match

    @staticmethod
    def get_matchs(pattern: str, value: any) -> list:
        re_pattern = re.compile(pattern)
        return [match for match in re.finditer(re_pattern, value)]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import numpy as np

from msprobe.pytorch.api_accuracy_checker.compare.algorithm import check_inf_nan_value, check_norm_value, \
    check_small_value
from msprobe.pytorch.api_accuracy_checker.precision_standard.base_standard import BaseCompare
from msprobe.pytorch.api_accuracy_checker.precision_standard.standard_config import StandardConfig
from msprobe.core.common.const import CompareConst



class AbsolutethdCompare(BaseCompare):
    """
    Absolute threshold compare class.

    This class is used to compare the absolute threshold of benchmark outputs and device outputs.
    It calculates various metrics such as inf_nan_error_ratio, rel_err_ratio, and abs_err_ratio
    to determine the accuracy of the device output compared to the benchmark output.

    Attributes:
        bench_output (np.ndarray): The output from the benchmark.
        device_output (np.ndarray): The output from the device.
        dtype (torch.dtype): The data type of the outputs.
        abs_bench (np.ndarray): The absolute value of the benchmark output.
        abs_bench_with_eps (np.ndarray): The absolute value of the benchmark output with epsilon.
        both_finite_mask (np.ndarray): A mask indicating where both outputs are finite.
        inf_nan_mask (np.ndarray): A mask indicating where either output is infinite or NaN.
        rtol (float): The relative tolerance for comparison.
        rel_err (np.ndarray): The relative error between the benchmark and device outputs.
        small_value (float): The small value threshold for comparison.
        small_value_atol (float): The absolute tolerance for small values.
        small_value_mask (np.ndarray): A mask indicating where values are small.
        normal_value_mask (np.ndarray): A mask indicating where values are normal.

    Methods:
        _get_rtol(): Gets the relative tolerance based on the data type.
        _get_rel_err(abs_bench_with_eps): Calculates the relative error.
        _get_normal_value_mask(small_value_mask): Gets the mask for normal values.
        _pre_compare(): Prepares the comparison by calculating various metrics.
        _compute_metrics(): Computes the comparison metrics.

    Note:
        This class assumes that the input data is a dictionary containing 'bench_output', 'device_output', 
        'compare_column' and 'dtype'.
        The 'dtype' should be a PyTorch data type.

    See Also:
        BaseCompare: The base class for comparison classes.
        StandardConfig: The class containing standard configuration values.
    """
    def __init__(self, input_data):
        super(AbsolutethdCompare, self).__init__(input_data)
        self.compare_algorithm = CompareConst.ABSOLUTE_THRESHOLD

    def _get_rtol(self):
        return StandardConfig.get_rtol(self.dtype)

    def _pre_compare(self):
        """
        Prepares the comparison by calculating various metrics.

        This method performs the following steps:
            1. Calculates the absolute benchmark values and their epsilon-adjusted versions.
            2. Determines masks for finite and infinite/NaN values in the outputs.
            3. Computes the absolute error between benchmark and device outputs.
            4. Retrieves the relative tolerance based on the data type.
            5. Calculates the relative error using the absolute error and epsilon-adjusted benchmark values.
            6. Determines the small value threshold and its absolute tolerance.
            7. Creates a mask for small values based on the benchmark values and finite mask.
            8. Creates a mask for normal values by excluding small values from the finite mask.
        """
        self.abs_bench, self.abs_bench_with_eps = self.stat_abs_bench_with_eps()
        self.both_finite_mask, self.inf_nan_mask = self.stat_finite_and_infinite_mask()
        self.abs_err = self.stat_abs_error()
        self.rtol = self._get_rtol()
        self.rel_err = self._get_rel_err(self.abs_err, self.abs_bench_with_eps)
        self.small_value, self.small_value_atol = self.get_small_value_threshold()
        self.small_value_mask = self.stat_small_value_mask(self.abs_bench, self.both_finite_mask, self.small_value)
        self.normal_value_mask = self._get_normal_value_mask(self.both_finite_mask, self.small_value_mask)

    def _compute_metrics(self):
        inf_nan_error_ratio = check_inf_nan_value(self.inf_nan_mask, self.bench_output, self.device_output, self.dtype,
                                                  self.rtol)
        rel_err_ratio = check_norm_value(self.normal_value_mask, self.rel_err, self.rtol)
        abs_err_ratio = check_small_value(self.abs_err, self.small_value_mask, self.small_value_atol)
        return {
            "inf_nan_error_ratio": inf_nan_error_ratio,
            "rel_err_ratio": rel_err_ratio,
            "abs_err_ratio": abs_err_ratio
        }



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import argparse
import os

from msprobe.core.common.file_utils import check_file_or_directory_path, create_directory
from msprobe.core.common.utils import Const, MsprobeBaseException


class UniqueDeviceAction(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        unique_values = set(values)
        if len(values) != len(unique_values):
            parser.error("device id must be unique")
        for device_id in values:
            if not 0 <= device_id <= 4095:
                parser.error(f"the argument 'device_id' must be in range [0, 4095], but got {device_id}")
        setattr(namespace, self.dest, values)


def add_api_accuracy_checker_argument(parser):
    parser.add_argument("-api_info", "--api_info_file", dest="api_info_file", type=str, required=True,
                        help="<Required> The api param tool result file: generate from api param tool, "
                             "a json file.")
    parser.add_argument("-o", "--out_path", dest="out_path", default="./", type=str, required=False,
                        help="<optional> The acc_check task result out path.")
    parser.add_argument("-csv_path", "--result_csv_path", dest="result_csv_path", default="", type=str, required=False,
                        help="<optional> the exit csv for continue")
    parser.add_argument('-save_error_data', dest="save_error_data", action="store_true",
                        help="<optional> Save compare failed api output.", required=False)


def multi_add_api_accuracy_checker_argument(parser):
    parser.add_argument("-api_info", "--api_info_file", dest="api_info_file", type=str, required=True,
                        help="<Required> The api param tool result file: generate from api param tool, "
                             "a json file.")
    parser.add_argument("-o", "--out_path", dest="out_path", default="./", type=str, required=False,
                        help="<optional> The acc_check task result out path.")
    parser.add_argument("-csv_path", "--result_csv_path", dest="result_csv_path", default="", type=str, required=False,
                        help="<optional> the exit csv for continue")
    parser.add_argument('-save_error_data', dest="save_error_data", action="store_true",
                        help="<optional> Save compare failed api output.", required=False)
    #以下属于多线程参数
    parser.add_argument("-d", "--device", dest="device_id", nargs='+', type=int,
                        help="<optional> set device id to acc_check, must be unique and in range 0-7",
                        default=[0], required=False, action=UniqueDeviceAction)


def check_args(args):
    args.api_info_file = os.path.abspath(args.api_info_file)
    check_file_or_directory_path(args.api_info_file)

    if args.out_path == "":
        args.out_path = "./"
    args.out_path = os.path.abspath(args.out_path)
    create_directory(args.out_path)

    if args.result_csv_path:
        args.result_csv_path = os.path.abspath(args.result_csv_path)
        check_file_or_directory_path(args.result_csv_path)



# coding=utf-8
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
Function:
TensorConversion class. This class mainly involves the convert_shape function.
"""

from msprobe.msaccucmp.dump_parse.dump import CompareData
from msprobe.msaccucmp.format_manager.format_manager import FormatManager
from msprobe.msaccucmp.format_manager.format_manager import SrcToDest
from msprobe.msaccucmp.format_manager.format_manager import ShapeConversion
from msprobe.msaccucmp.vector_cmp.fusion_manager.fusion_op import FusionOp
from msprobe.msaccucmp.vector_cmp.fusion_manager.fusion_op import Tensor
from msprobe.msaccucmp.cmp_utils import log, utils_type
from msprobe.msaccucmp.cmp_utils import utils
from msprobe.msaccucmp.cmp_utils import common
from msprobe.msaccucmp.cmp_utils.constant.const_manager import ConstManager, DD
from msprobe.msaccucmp.cmp_utils.constant.compare_error import CompareError


class TensorConversion:
    """
    The class for tensor conversion
    """

    NEED_CONVERT_TYPE = (DD.FORMAT_FRACTAL_NZ, DD.FORMAT_NDC1HWC0)
    ND = DD.FORMAT_ND

    def __init__(self: any, fusion_op: FusionOp, format_manager: FormatManager, is_detail: bool) -> None:
        self.fusion_op = fusion_op
        self.shape_conversion = ShapeConversion(format_manager)
        self.is_detail = is_detail

    @staticmethod
    def _translate_shape_to_array(shape_tuple: list) -> (list, bool):
        all_shape_is_1 = False
        remove_1_shape = []
        for item in shape_tuple:
            if item != 1:
                remove_1_shape.append(item)
        if len(remove_1_shape) == 0:
            return shape_tuple, True
        return remove_1_shape, all_shape_is_1

    @staticmethod
    def _padding_shape(origin_shape: list, expect_length: int) -> (list, int, int):
        shape = []
        size = 1
        one_count = 0
        for dim in origin_shape:
            shape.append(dim)
            size *= dim
            if dim == 1:
                one_count += 1
        if expect_length > len(shape):
            left_count = expect_length - len(shape)
            for _ in range(left_count):
                shape.append(1)
        return shape, size, one_count

    @staticmethod
    def _change_format(ground_truth_tensor: Tensor, origin_shape: any) -> None:
        # if fusion rule is (1,224,224, 3),origin is (1,3, 224,224),
        is_original_shape_nhwc = ground_truth_tensor.shape[0] == origin_shape[0] and \
                                 ground_truth_tensor.shape[1] == origin_shape[2] and \
                                 ground_truth_tensor.shape[2] == origin_shape[3] and \
                                 ground_truth_tensor.shape[3] == origin_shape[1]
        # need change format to NCHW
        if ground_truth_tensor.tensor_format == 'NHWC' and is_original_shape_nhwc:
            ground_truth_tensor.tensor_format = 'NCHW'
        # if fusion rule is (1,3,224,224),origin is (1,224,224,3),
        is_original_shape_nchw = ground_truth_tensor.shape[0] == origin_shape[0] and \
                                 ground_truth_tensor.shape[1] == origin_shape[3] and \
                                 ground_truth_tensor.shape[2] == origin_shape[1] and \
                                 ground_truth_tensor.shape[3] == origin_shape[2]
        # need change format to NHWC
        if ground_truth_tensor.tensor_format == 'NCHW' and is_original_shape_nchw:
            ground_truth_tensor.tensor_format = 'NHWC'

    @staticmethod
    def _make_detail_dest_format(my_output_tensor: any, ground_truth_format: int) -> (any, any):
        # convert my output and ground truth format to nchw
        if common.contain_depth_dimension(my_output_tensor.tensor_format):
            my_output_dest_format = DD.FORMAT_NCDHW
            ground_truth_dest_format = DD.FORMAT_NCDHW
        else:
            if my_output_tensor.tensor_format == DD.FORMAT_FRACTAL_Z:
                my_output_dest_format = ground_truth_format
                ground_truth_dest_format = ground_truth_format
            else:
                my_output_dest_format = DD.FORMAT_NCHW
                ground_truth_dest_format = DD.FORMAT_NCHW
        return my_output_dest_format, ground_truth_dest_format

    def slice_data(self: any, my_output_np: any, adjusted_shape: list) -> any:
        """
        Slice data by right shape
        :param my_output_np: my output numpy data
        :param adjusted_shape: the shape to slice
        :return the sliced numpy data
        """
        old_my_output_shape_str = utils.convert_shape_to_string(my_output_np.shape)
        if len(my_output_np.shape) < len(adjusted_shape):
            self._check_ground_truth_shape_by_my_output_shape(my_output_np.shape, adjusted_shape)
            return my_output_np
        # padding to 4d
        my_output_shape, my_output_size, one_count = self._padding_shape(
            my_output_np.shape, ConstManager.FOUR_DIMS_LENGTH)

        # padding to my output shape
        slice_shape, ground_truth_size, _ = self._padding_shape(adjusted_shape, len(my_output_shape))

        if ground_truth_size == my_output_size:
            if one_count == len(my_output_np.shape) or one_count == len(my_output_np.shape) - 1:
                return my_output_np

        # my output shape large then ground truth dump data, need slice data
        if self._should_slice(my_output_shape, slice_shape):
            # slice data
            my_output_np = my_output_np.reshape(my_output_shape)
            if len(my_output_shape) == ConstManager.FOUR_DIMS_LENGTH:
                my_output_np = my_output_np[:slice_shape[0], :slice_shape[1], :slice_shape[2], :slice_shape[3]]
            elif len(my_output_shape) == ConstManager.FIVE_DIMS_LENGTH:
                my_output_np = \
                    my_output_np[:slice_shape[0], :slice_shape[1], :slice_shape[2], :slice_shape[3], :slice_shape[4]]
            if self.fusion_op:
                log.print_info_log('[%s] The left dump data has been sliced from %s to %s.' %
                                   (self.fusion_op.op_name, old_my_output_shape_str,
                                    utils.convert_shape_to_string(slice_shape)))
        return my_output_np

    def get_my_output_and_ground_truth_data(self: any, compare_data: CompareData, my_output_tensor: any,
                                            ground_truth_tensor: Tensor) -> (any, any, any):
        """
        Deserialize the my output and ground truth tensor to  array
        :param compare_data: the compare data
        :param my_output_tensor: my output tensor
        :param ground_truth_tensor: ground truth tensor
        :return: my_output_array, ground_truth_array, shape
        """
        # get ground truth format
        origin_format = my_output_tensor.tensor_format
        if ground_truth_tensor.tensor_format != "":
            origin_format = self._get_ground_truth_format(ground_truth_tensor)
        is_tensor = (utils.get_shape_type(my_output_tensor.shape) == utils_type.ShapeType.Tensor)
        my_output_shape = utils.convert_shape_to_string(my_output_tensor.shape)

        # when compare quant and origin
        if compare_data.is_standard_quant_vs_origin():
            # deserialize dump data to np array
            my_output_array = my_output_tensor.data
            ground_truth_array = ground_truth_tensor.data.data
            log.print_info_log('[%s] Left %s <======> Right %s'
                               % (self.fusion_op.op_name, my_output_shape,
                                  utils.convert_shape_to_string(ground_truth_tensor.data.shape)))
            self._check_shape_valid(my_output_tensor.shape, ground_truth_tensor.data.shape)
            return my_output_array, ground_truth_array, my_output_tensor.shape

        log.print_info_log('[%s] Before shape convert, Left %s%s <======> Right %s%s'
                           % (self.fusion_op.op_name, common.get_format_string(my_output_tensor.tensor_format),
                              my_output_shape, ground_truth_tensor.tensor_format,
                              utils.convert_shape_to_string(ground_truth_tensor.data.shape)))
        # convert shape
        my_output_np, ground_truth_np = self._convert_shape(my_output_tensor, ground_truth_tensor, origin_format)

        # slice data
        if (my_output_tensor.tensor_format != self.ND or
           (is_tensor and (my_output_np.size != ground_truth_np.size))):
            my_output_np = self.slice_data(my_output_np, ground_truth_np.shape)

        my_output_array = my_output_np.flatten()
        ground_truth_array = ground_truth_np.flatten()
        if len(my_output_array) != len(ground_truth_array):
            message = log.print_cannot_compare_warning(
                self.fusion_op.op_name, '(%d)' % len(my_output_array), '(%d)' % len(ground_truth_array))
            raise CompareError(CompareError.MSACCUCMP_INVALID_SHAPE_ERROR, message)
        return my_output_array, ground_truth_array, my_output_np.shape

    def _check_ground_truth_shape_by_my_output_shape(self: any, my_output_shape: any,
                                                     ground_truth_shape: any) -> None:
        my_output_shape_array, my_output_all_shape_is_1 = self._translate_shape_to_array(my_output_shape)
        ground_truth_shape_array, ground_truth_all_shape_is_1 = self._translate_shape_to_array(ground_truth_shape)
        if my_output_all_shape_is_1 and ground_truth_all_shape_is_1:
            return
        self._check_shape_valid(my_output_shape_array, ground_truth_shape_array)

    def _should_slice(self: any, my_output_shape: list, slice_shape: list) -> bool:
        is_slice = False
        for my_output_dim, ground_truth_dim in zip(my_output_shape, slice_shape):
            if my_output_dim < ground_truth_dim:
                if self.fusion_op:
                    old_my_output_shape_str = utils.convert_shape_to_string(my_output_shape)
                    slice_shape_str = utils.convert_shape_to_string(slice_shape)
                    message = log.print_cannot_compare_warning(self.fusion_op.op_name, old_my_output_shape_str,
                                                               slice_shape_str)
                    raise CompareError(CompareError.MSACCUCMP_INVALID_SHAPE_ERROR, message)
                raise CompareError(CompareError.MSACCUCMP_INVALID_SHAPE_ERROR)
            if my_output_dim > ground_truth_dim:
                is_slice = True
                break
        return is_slice

    def _change_format_by_origin_shape(self: any, ground_truth_tensor: Tensor) -> None:
        origin_shape = ground_truth_tensor.data.shape
        if len(ground_truth_tensor.shape) == ConstManager.FOUR_DIMS_LENGTH and \
                len(origin_shape) == ConstManager.FOUR_DIMS_LENGTH:
            match = True
            for (ground_truth_dim, origin_dim) in zip(ground_truth_tensor.shape, origin_shape):
                if ground_truth_dim != origin_dim:
                    match = False
                    break
            if not match:
                self._change_format(ground_truth_tensor, origin_shape)

    def _get_ground_truth_format(self: any, ground_truth_tensor: Tensor) -> int:
        if ground_truth_tensor.tensor_format not in ConstManager.STRING_TO_FORMAT_MAP:
            message = 'The format " %s " is not supported.' % ground_truth_tensor.tensor_format
            log.print_warn_log(message)
            raise CompareError(CompareError.MSACCUCMP_INVALID_FORMAT_ERROR, message)
        if ground_truth_tensor.tensor_format in ('NHWC', 'NCHW'):
            self._change_format_by_origin_shape(ground_truth_tensor)
        return ConstManager.STRING_TO_FORMAT_MAP.get(ground_truth_tensor.tensor_format)

    def _check_shape_valid(self: any, my_output_shape: any, ground_truth_shape: any) -> None:
        if len(my_output_shape) != len(ground_truth_shape):
            op_name = self.fusion_op.op_name if self.fusion_op else ''
            message = log.print_cannot_compare_warning(
                op_name, utils.convert_shape_to_string(my_output_shape),
                utils.convert_shape_to_string(ground_truth_shape))
            raise CompareError(CompareError.MSACCUCMP_INVALID_SHAPE_ERROR, message)
        for (my_output_dim, ground_truth_dim) in zip(my_output_shape, ground_truth_shape):
            if my_output_dim != ground_truth_dim:
                if self.fusion_op:
                    message = log.print_cannot_compare_warning(self.fusion_op.op_name,
                                                               utils.convert_shape_to_string(my_output_shape),
                                                               utils.convert_shape_to_string(ground_truth_shape))
                    raise CompareError(CompareError.MSACCUCMP_INVALID_SHAPE_ERROR, message)
                raise CompareError(CompareError.MSACCUCMP_INVALID_SHAPE_ERROR)

    def _convert_shape(self: any, my_output_tensor: any, ground_truth_tensor: Tensor,
                       ground_truth_format: int) -> (any, any):
        my_output_dest_format = my_output_tensor.tensor_format
        ground_truth_dest_format = ground_truth_format
        if self.is_detail:
            # convert my output and ground truth format to nchw
            my_output_dest_format, ground_truth_dest_format = self._make_detail_dest_format(
                my_output_tensor, ground_truth_format)
        else:
            if my_output_tensor.tensor_format != ground_truth_tensor.tensor_format:
                my_output_dest_format = ground_truth_format

        # deserialize dump data to np array
        my_output_array = my_output_tensor.data
        ground_truth_array = ground_truth_tensor.data.data

        # ND format no need to convert, except (FORMAT_FRACTAL_NZ or NDC1HWC0) to ND
        if my_output_dest_format == self.ND and my_output_tensor.tensor_format not in self.NEED_CONVERT_TYPE \
                or utils.get_shape_type(my_output_tensor.shape) != utils_type.ShapeType.Tensor:
            return my_output_array, ground_truth_array

        # shape convert for my output
        my_output_group = common.get_sub_format(my_output_tensor)
        my_output_src_to_dest = SrcToDest(my_output_tensor.tensor_format, my_output_dest_format, my_output_tensor.shape,
                                          ground_truth_tensor.shape)
        my_output_np = self.shape_conversion.convert_shape(my_output_src_to_dest, my_output_array,
                                                           {'group': my_output_group})

        # shape convert for ground truth
        ground_truth_group = common.get_sub_format(ground_truth_tensor, my_output_group)
        ground_truth_src_to_dest = SrcToDest(ground_truth_format, ground_truth_dest_format,
                                             ground_truth_tensor.data.shape,
                                             ground_truth_tensor.data.shape)
        ground_truth_np = self.shape_conversion.convert_shape(ground_truth_src_to_dest, ground_truth_array,
                                                              {'group': ground_truth_group})

        log.print_info_log('[%s] After shape convert, Left %s%s <======> Right %s%s'
                           % (self.fusion_op.op_name, common.get_format_string(my_output_dest_format),
                              str(my_output_np.shape), common.get_format_string(ground_truth_dest_format),
                              str(ground_truth_np.shape)))
        return my_output_np, ground_truth_np


class ConvertSingleTensorFormat:
    def __init__(self, custom_path: str = "", additional_target_dim_to_format: dict = None) -> None:
        if custom_path and not isinstance(custom_path, str):
            raise CompareError(CompareError.MSACCUCMP_INVALID_PARAM_ERROR, "custom_path needs to be a string")

        self.manager = FormatManager(custom_path=custom_path)
        self.manager.check_arguments_valid()
        self.shape_conversion = ShapeConversion(self.manager)
        self.tensor_conversion = TensorConversion(fusion_op=None, format_manager=self.manager, is_detail=False)

        self.nd_source_formats = [ConstManager.STRING_TO_FORMAT_MAP.get(ii, None) for ii in ['HWCN', 'NCHW', 'NHWC']]
        self.nd_target_format = ConstManager.STRING_TO_FORMAT_MAP.get("ND", None)
        if self.nd_target_format is None:
            message = "This should not happen, check if the installed CANN toolkit is broken"
            raise CompareError(CompareError.MSACCUCMP_INVALID_PARAM_ERROR, message)

        self.target_dim_to_format = {
            4: ConstManager.STRING_TO_FORMAT_MAP.get("NCHW"),
            3: self.nd_target_format,
            2: self.nd_target_format,
        }
        if additional_target_dim_to_format:
            self._check_and_set_additional_target_dim_to_format(additional_target_dim_to_format)
            self.target_dim_to_format.update(additional_target_dim_to_format)
        self.nchw_format_len = 4

    def __call__(self, my_tensor: Tensor) -> any:
        real_format = my_tensor.tensor_format
        if my_tensor.tensor_format in self.nd_source_formats:
            if len(my_tensor.shape) != self.nchw_format_len:
                real_format = self.nd_target_format
        shape_to = my_tensor.original_shape
        format_to = self.target_dim_to_format.get(len(shape_to), real_format)
        src_to_dest = SrcToDest(real_format, format_to, my_tensor.shape, shape_to)

        if real_format == format_to:
            dump_data_np = self.shape_conversion.reshape(src_to_dest, my_tensor.data)
        else:
            try:
                dump_data_np = self.manager.execute_format_convert(
                  src_to_dest, my_tensor.data, {'group': common.get_sub_format(my_tensor)}
                )
            except CompareError as ee:
                log.print_error_log(ee)
                return my_tensor.data.reshape(my_tensor.shape)

        # slice data
        if len(shape_to) > 0 and real_format != format_to:
            try:
                dump_data_np = self.tensor_conversion.slice_data(dump_data_np, shape_to)
            except CompareError as ee:
                log.print_error_log(ee)
                return dump_data_np
        return dump_data_np

    @staticmethod
    def _check_and_set_additional_target_dim_to_format(additional_target_dim_to_format):
        if additional_target_dim_to_format is None:
            return
        if not isinstance(additional_target_dim_to_format, dict):
            message = "additional_target_dim_to_format needs to be None or a dict"
            raise CompareError(CompareError.MSACCUCMP_INVALID_PARAM_ERROR, message)

        for source_dim, target_format in list(additional_target_dim_to_format.items()):  # like {4: "NCHW", 5: "ND"}
            if not isinstance(source_dim, int):
                message = "additional_target_dim_to_format key should be an int value indicates source dim"
                raise CompareError(CompareError.MSACCUCMP_INVALID_PARAM_ERROR, message)
            if not isinstance(target_format, str) or target_format not in ConstManager.STRING_TO_FORMAT_MAP:
                message = "additional_target_dim_to_format value should be a string indicates target format like NCHW"
                raise CompareError(CompareError.MSACCUCMP_INVALID_PARAM_ERROR, message)
            additional_target_dim_to_format[source_dim] = ConstManager.STRING_TO_FORMAT_MAP.get(target_format)




# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import gc
import os
import threading
from abc import ABC, abstractmethod
from collections import defaultdict

from msprobe.core.common.runtime import Runtime
from msprobe.core.common.utils import Const, ThreadSafe
from msprobe.core.dump.data_dump.data_processor.base import (ModuleBackwardInputsOutputs, ModuleForwardInputsOutputs)


class HookSet:
    def __init__(
            self,
            forward_pre_hook=None,
            forward_hook=None,
            backward_pre_hook=None,
            backward_hook=None,
            distributed_forward_hook=None
    ):
        self.forward_pre_hook = forward_pre_hook
        self.forward_hook = forward_hook
        self.backward_pre_hook = backward_pre_hook
        self.backward_hook = backward_hook
        self.distributed_forward_hook = distributed_forward_hook


class BaseHookManager(ABC):
    inner_switch = defaultdict(bool)
    inner_api_count = defaultdict(int)
    hook_handle_dict = {}
    params_grad_info = {}
    grad_hook_call = {}

    def __init__(self, data_collector, config):
        self.data_collector = data_collector
        self.config = config
        self.current_pid = self._pid

    @property
    def _pid(self):
        return os.getpid()

    @staticmethod
    def reset_status():
        BaseHookManager.inner_switch = defaultdict(bool)
        BaseHookManager.inner_api_count = defaultdict(int)
        BaseHookManager.params_grad_info.clear()

    @staticmethod
    def ensure_gc_enabled():
        is_gc_disabled = not gc.isenabled()
        if is_gc_disabled:
            gc.enable()
        return is_gc_disabled

    @staticmethod
    def restore_gc_state(original_state):
        if original_state:
            gc.disable()

    @staticmethod
    def _clear_input_kwargs(module, tid):
        if hasattr(module, 'msprobe_input_kwargs') and tid in module.msprobe_input_kwargs:
            del module.msprobe_input_kwargs[tid]

    @staticmethod
    def _get_grad_hook_call_index(ori_name, param_name):
        if ori_name not in BaseHookManager.grad_hook_call:
            BaseHookManager.grad_hook_call[ori_name] = [0, param_name]
        else:
            if BaseHookManager.grad_hook_call.get(ori_name)[1] == param_name:
                BaseHookManager.grad_hook_call[ori_name][0] += 1
        return BaseHookManager.grad_hook_call.get(ori_name)[0]

    @staticmethod
    @abstractmethod
    def _no_grad_context():
        pass

    @staticmethod
    @abstractmethod
    def _add_count(name):
        pass

    @staticmethod
    @abstractmethod
    def _get_count(name):
        pass

    @staticmethod
    @abstractmethod
    def _process_kwargs_and_output(module, tid, hook_type, kwargs_or_output, output_or_kwargs):
        pass

    @abstractmethod
    def build_hook(self, hook_type, name):
        pass

    @abstractmethod
    def _register_forward_hook(self, module, api_name):
        pass

    @abstractmethod
    def _register_backward_hook(self, module, full_backward_name, args):
        pass

    @abstractmethod
    def _register_backward_pre_hook(self, module, full_backward_name, args, kwargs, output):
        pass

    @abstractmethod
    def _get_params_dict(self, module):
        pass

    @abstractmethod
    def _need_exchange(self, module):
        pass

    @abstractmethod
    def _register_param_hook(self, name, module, params_dict):
        pass

    def is_child_process(self):
        return self.current_pid != self._pid

    def _should_execute_hook(self, hook_type, tid, is_forward=True):
        is_api_hook = hook_type == Const.API
        if self.is_child_process():
            return False
        if BaseHookManager.inner_switch[tid]:
            return False
        if not is_api_hook and not Runtime.is_running:
            return False
        elif is_api_hook and is_forward and not Runtime.is_running:
            return False
        if not self.data_collector or self.data_collector.data_processor.is_terminated:
            return False
        return True

    def _build_forward_pre_hook(self, hook_type, api_name):
        def forward_pre_hook(module, args, kwargs=None):
            if hook_type == Const.MODULE:
                return None

            tid = threading.get_ident()
            if not self._should_execute_hook(hook_type, tid):
                return None

            with ThreadSafe():
                original_state = self.ensure_gc_enabled()
                self._register_forward_hook(module, api_name)
                BaseHookManager.inner_api_count[tid] += 1
                if BaseHookManager.inner_api_count[tid] != 1:
                    return None

                full_forward_name = api_name + str(self._get_count(api_name)) + Const.SEP + Const.FORWARD
                full_backward_name = api_name + str(self._get_count(api_name)) + Const.SEP + Const.BACKWARD
                module.full_forward_name = full_forward_name
                if kwargs is None:
                    kwargs = module.msprobe_input_kwargs.get(tid, {}) if hasattr(module, 'msprobe_input_kwargs') else {}
                BaseHookManager.inner_switch[tid] = True
                module_input_output = ModuleForwardInputsOutputs(args=args, kwargs=kwargs, output=None)

                args = self._register_backward_hook(module, full_backward_name, args)
                with self._no_grad_context():
                    self.data_collector.update_api_or_module_name(full_forward_name)
                    self.data_collector.forward_input_data_collect(
                        full_forward_name,
                        module,
                        self._pid,
                        module_input_output
                    )
                BaseHookManager.inner_switch[tid] = False
                self.restore_gc_state(original_state)
                return args

        return forward_pre_hook

    def _build_forward_hook(self, hook_type, api_name):
        def forward_hook(module, args, kwargs_or_output, output_or_kwargs=None):
            tid = threading.get_ident()
            if not self._should_execute_hook(hook_type, tid):
                self._clear_input_kwargs(module, tid)
                return None

            with ThreadSafe():
                original_state = self.ensure_gc_enabled()
                if hook_type == Const.API:
                    if BaseHookManager.inner_api_count[tid] != 1:
                        if BaseHookManager.inner_api_count[tid] > 1:
                            BaseHookManager.inner_api_count[tid] -= 1
                        self._clear_input_kwargs(module, tid)
                        return None

                kwargs, output = self._process_kwargs_and_output(
                    module,
                    tid,
                    hook_type,
                    kwargs_or_output,
                    output_or_kwargs
                )
                BaseHookManager.inner_switch[tid] = True
                module_input_output = ModuleForwardInputsOutputs(args=args, kwargs=kwargs, output=output)
                if hook_type == Const.API:
                    full_forward_name = api_name + str(self._get_count(api_name)) + Const.SEP + Const.FORWARD
                    full_backward_name = api_name + str(self._get_count(api_name)) + Const.SEP + Const.BACKWARD
                    output = self._register_backward_pre_hook(module, full_backward_name, args, kwargs, output)

                with self._no_grad_context():
                    if hook_type == Const.MODULE:
                        params_dict = self._get_params_dict(module)
                        setattr(module_input_output, Const.PARAMS, params_dict)
                        if params_dict:
                            self._register_param_hook(api_name, module, params_dict)
                        self.data_collector.update_api_or_module_name(api_name)
                        self.data_collector.forward_data_collect(
                            api_name,
                            module,
                            self._pid,
                            module_input_output
                        )
                    else:
                        self.data_collector.update_api_or_module_name(full_forward_name)
                        self.data_collector.forward_output_data_collect(
                            full_forward_name,
                            module,
                            self._pid,
                            module_input_output
                        )
                        self._add_count(api_name)
                        BaseHookManager.inner_api_count[tid] -= 1
                    self._clear_input_kwargs(module, tid)
                BaseHookManager.inner_switch[tid] = False
                self.restore_gc_state(original_state)
                return output

        return forward_hook

    def _build_backward_hook(self, hook_type, full_name):
        def backward_hook(module, grad_input, grad_output):
            tid = threading.get_ident()
            if not self._should_execute_hook(hook_type, tid, is_forward=False):
                return

            with ThreadSafe():
                original_state = self.ensure_gc_enabled()
                BaseHookManager.inner_switch[tid] = True
                self.data_collector.update_api_or_module_name(full_name)

                need_exchange = self._need_exchange(module) if hook_type == Const.MODULE else True
                if need_exchange:
                    module_input_output = ModuleBackwardInputsOutputs(grad_input=grad_output, grad_output=grad_input)
                else:
                    module_input_output = ModuleBackwardInputsOutputs(grad_input=grad_input, grad_output=grad_output)
                self.data_collector.backward_data_collect(
                    full_name,
                    module,
                    self._pid,
                    module_input_output
                )
                BaseHookManager.inner_switch[tid] = False
                self.restore_gc_state(original_state)

        return backward_hook



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from msprobe.core.common.const import Const
from msprobe.mindspore.dump.debugger.debugger_config import DebuggerConfig
from msprobe.mindspore.dump.dump_processor.dump_tool_factory import DumpToolFactory
from msprobe.mindspore.dump.overflow_check.overflow_check_tool_factory import OverflowCheckToolFactory
from msprobe.mindspore.dump.exception_dump.exception_dump_tool_factory import ExceptionDumpToolFactory


class TaskHandlerFactory:
    tasks = {
        Const.TENSOR: DumpToolFactory,
        Const.STATISTICS: DumpToolFactory,
        Const.OVERFLOW_CHECK: OverflowCheckToolFactory,
        Const.EXCEPTION_DUMP: ExceptionDumpToolFactory
    }

    @staticmethod
    def create(config: DebuggerConfig, model=None):
        task = TaskHandlerFactory.tasks.get(config.task)
        if not task:
            raise Exception("Valid task is needed.")
        if task == DumpToolFactory:
            handler = task.create(config, model)
        else:
            handler = task.create(config)
        if not handler:
            raise Exception("Can not find task handler")
        return handler



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from dataclasses import dataclass
from msprobe.core.common.const import Const
from msprobe.core.common.log import logger
from msprobe.overflow_check.utils import FileCache, RankPath, is_ignore_op, check_item_anomaly, OverFlowCheckConst
from msprobe.core.common.exceptions import MsprobeException


@dataclass
class DataNode:
    op_name: str
    rank: int
    inputs: list
    input_args: list
    input_kwargs: dict
    outputs: dict
    layer: int = 0  # 和communication_node的layer保持一致
    sub_layer: int = 0  # 调用顺序，越小表示越先调用

    def __init__(self, op_name, rank, op_data, **kwargs):
        self.op_name = op_name
        self.rank = rank
        self.inputs = op_data.get(Const.INPUT, [])
        self.input_args = op_data.get(Const.INPUT_ARGS, [])
        self.input_kwargs = op_data.get(Const.INPUT_KWARGS, {})
        self.outputs = op_data.get(Const.OUTPUT, {})
        self.sub_layer = kwargs.get('sub_layer', 0)

    @staticmethod
    def find_complete_construct(construct_info, op_name):
        construct = [op_name]
        seen = set(op_name)
        while True:
            op_name = construct_info.get(op_name)
            if isinstance(op_name, list):
                op_name = op_name[0]
            if not op_name or op_name in seen:
                return construct
            construct.insert(0, op_name)
            seen.add(op_name)

    def find_stack(self, stack_info):
        for item in stack_info.values():
            if not isinstance(item, list):
                raise MsprobeException(MsprobeException.UNSUPPORTED_TYPE_ERROR,
                                       f'The value\'s type in stack.json should be a list, not {type(item)}!')
            if len(item) >= 2 and self.op_name in item[0]:
                return item[1]
        return {}

    def is_anomaly(self) -> bool:
        if is_ignore_op(self.op_name):
            return False
        is_input_anomaly = (check_item_anomaly(self.inputs) or check_item_anomaly(self.input_args) or
                            check_item_anomaly(self.input_kwargs))
        is_output_anomaly = check_item_anomaly(self.outputs)
        return (not is_input_anomaly) and is_output_anomaly

    def gen_node_info(self, path: RankPath):
        cache = FileCache()
        construct = cache.load_json(path.construct_path)
        stack = cache.load_json(path.stack_path)
        if Const.FORWARD in self.op_name:
            data_info_list = {Const.INPUT_ARGS: self.input_args, Const.INPUT_KWARGS: self.input_kwargs,
                              Const.OUTPUT: self.outputs}
        else:
            data_info_list = {Const.INPUT: self.inputs, Const.OUTPUT: self.outputs}
        return {'op_name': self.op_name,
                'data_info': data_info_list,
                'construct_info': self.find_complete_construct(construct, self.op_name),
                'stack_info': self.find_stack(stack)}


class CommunicationNode:
    def __init__(self, node_id, rank, data: DataNode, layer=0, **kwargs):
        self.node_id = node_id
        self.rank = rank
        self.data = data
        self.layer = layer
        op_name_split = self.data.op_name.split(Const.SEP)
        if len(op_name_split) < 4:
            logger.error(f'invalid op_name: {self.data.op_name}')
            raise RuntimeError(f'invalid op_name: {self.data.op_name}')
        self.api = op_name_split[1]
        self.call_cnt = op_name_split[2]
        self.pre_node = kwargs.get('pre_node')
        self.link_nodes = kwargs.get('link_nodes', {})
        self.dst_nodes = kwargs.get('dst_nodes', {})
        self.src_nodes = kwargs.get('src_nodes', {})
        self.next_nodes = kwargs.get('next_nodes', {})
        self.compute_ops = kwargs.get('compute_ops', [])
        self.type = self._resolve_type()
        self.connected = False

    def add_next(self, node):
        self.next_nodes[node.node_id] = node
        node.pre_node = self
        node.layer = self.layer + 1
        node.data.layer = node.layer

    def add_link(self, node):
        self.link_nodes[node.node_id] = node
        node.link_nodes[self.node_id] = self
        node.layer = self.layer
        node.data.layer = node.layer
        self.connected = True
        node.connected = True

    def add_dst(self, node):
        self.dst_nodes[node.node_id] = node
        node.src_nodes[self.node_id] = self
        node.layer = self.layer
        node.data.layer = node.layer
        self.connected = True
        node.connected = True

    def delete(self):
        for node in self.next_nodes.values():
            node.pre_node = None
        for node in self.dst_nodes.values():
            node.src_nodes.pop(self.node_id)
        for node in self.src_nodes.values():
            node.dst_nodes.pop(self.node_id)
        for node in self.link_nodes.values():
            node.link_nodes.pop(self.node_id)
        if self.pre_node:
            self.pre_node.next_nodes.pop(self.node_id)

    def has_nan_inf(self):
        return self.input_has_nan_inf() or check_item_anomaly(self.data.outputs)

    def input_has_nan_inf(self):
        return check_item_anomaly(self.data.input_args) or check_item_anomaly(self.data.input_kwargs)

    def find_connected_nodes(self):
        """
        根据 api/类型/入参/调用次数 确定相连接的node的op_name
        """
        tar_api = OverFlowCheckConst.P2P_API_MAPPING.get(self.api, self.api)
        ranks = set()
        for dst in [OverFlowCheckConst.DST, OverFlowCheckConst.DST_GROUP]:
            if dst in self.data.input_kwargs:
                dst_value = self.data.input_kwargs.get(dst)
                if dst_value:
                    ranks.add(dst_value.get('value'))
                break
        for src in [OverFlowCheckConst.SRC, OverFlowCheckConst.SRC_GROUP]:
            if src in self.data.input_kwargs:
                src_value = self.data.input_kwargs.get(src)
                if src_value:
                    ranks.add(src_value.get('value'))
                break
        if not ranks:
            for item in self.data.input_args:
                if isinstance(item, dict) and item.get(Const.TYPE) == 'int':
                    ranks.add(item.get('value'))
        group = self.data.input_kwargs.get('group')
        if group:
            ranks.update(group.get('group_ranks'))
        return {'ranks': ranks, 'api': f'Distributed.{tar_api}',
                'type': OverFlowCheckConst.OPPOSITE_DIR.get(self.type, OverFlowCheckConst.LINK)}

    def _resolve_type(self):
        for src in [OverFlowCheckConst.SRC, OverFlowCheckConst.SRC_GROUP]:
            if src in self.data.input_kwargs and self.data.input_kwargs[src]:
                if self.data.input_kwargs[src].get('value') == self.rank:
                    return OverFlowCheckConst.SRC
                else:
                    return OverFlowCheckConst.DST
        for dst in [OverFlowCheckConst.DST, OverFlowCheckConst.DST_GROUP]:
            if dst in self.data.input_kwargs and self.data.input_kwargs[dst]:
                if self.data.input_kwargs[dst].get('value') == self.rank:
                    return OverFlowCheckConst.DST
                else:
                    return OverFlowCheckConst.SRC
        if self.api in OverFlowCheckConst.DIRECTED_API:
            for item in self.data.input_args:
                if item.get(Const.TYPE) == 'int':
                    node_type = OverFlowCheckConst.DIRECTED_API[self.api]
                    return node_type if item.get('value') == self.rank else OverFlowCheckConst.OPPOSITE_DIR[node_type]
        return OverFlowCheckConst.LINK



# coding=utf-8
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
Function:
Progress class. This class mainly involves the print_progress function.
"""
import time
import math

from msprobe.msaccucmp.cmp_utils import log
from msprobe.msaccucmp.cmp_utils.constant.const_manager import ConstManager
from msprobe.msaccucmp.cmp_utils.constant.compare_error import CompareError


class Progress:
    """
    The class for progress
    """
    PROGRESS_GREATER_THAN_COUNT = 50
    PROGRESS_GREATER_THAN = '>'
    INTERVAL_TIME_SECOND = 1

    def __init__(self: any, total_count: int) -> None:
        self.total_count = total_count
        self.last_progress_time = 0
        self.current_count = 0

    def update_progress(self: any, update_count: int = 1) -> None:
        """
        Update the progress
        """
        self.current_count += update_count

    def is_done(self: any) -> bool:
        """
        check if the process is done
        """
        return self.current_count == self.total_count

    def update_and_print_progress(self: any, progress: int = None) -> None:
        """
        Print the progress realtime
        :param progress: the progress
        """
        if progress is None:
            if self.total_count != 0:
                progress = round(self.current_count * 100.0 / self.total_count, 2)
            else:
                progress = 0
                log.print_error_log('Can not divide zero.')
        current_time = time.time()
        denominator = 1
        greater_than_count = 0
        if self.PROGRESS_GREATER_THAN_COUNT != 0:
            denominator = ConstManager.MAX_PROGRESS // self.PROGRESS_GREATER_THAN_COUNT
        if denominator != 0:
            greater_than_count = math.floor(progress / denominator)
        progress_info = '%s%s' % (self.PROGRESS_GREATER_THAN * greater_than_count,
                                  ' ' * (self.PROGRESS_GREATER_THAN_COUNT - greater_than_count))
        if current_time - self.last_progress_time >= self.INTERVAL_TIME_SECOND \
                or progress == ConstManager.MAX_PROGRESS:
            log.print_info_log('[ %s %d%%]' % (progress_info, progress))
            self.last_progress_time = current_time



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from typing import Any
from argparse import ArgumentTypeError
from msprobe.infer.utils.check.checker import Checker


class ArgsChecker:
    """
    parser.add_argument(..., type=ArgsChecker(Rule.to_int()), ...)
    """

    def __init__(self, rule: Checker) -> None:
        self.rule = rule

    def __call__(self, value) -> Any:
        try:
            self.rule.check(value, will_raise=True)
        except ValueError as err:
            raise ArgumentTypeError(str(err)) from err
        return self.rule.get_value()



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import json
from werkzeug import Response
from werkzeug.wrappers.request import Request
from ..utils.constant import security_headers
from ..utils.graph_utils import GraphUtils
from ..utils.global_state import GraphState


def check_file_type(func):
    def wrapper(*args, **kwargs):
        try:
            if len(args) <= 0:
                raise RuntimeError('Illegal function call, at least 1 parameter is required but got 0')
            request = args[0]
            if not isinstance(request, Request):
                raise RuntimeError('The request "parameter" is not in a format supported by werkzeug')
            data = GraphUtils.safe_json_loads(request.get_data().decode('utf-8'), {})
            meta_data = GraphUtils.safe_get_meta_data(data)
            result = {'success': False, 'error': ''}
            if meta_data is None or not isinstance(meta_data, dict):
                result['error'] = GraphUtils.t('metaDataError')
                return Response(json.dumps(result), content_type="application/json", headers=security_headers)
            # 设置语言
            GraphState.set_global_value('lang', meta_data.get('lang', 'zh-CN'))
        except Exception as e:
            result = {'success': False, 'error': str(e)}
            return Response(json.dumps(result), content_type="application/json", headers=security_headers)
        return func(*args, **kwargs)

    return wrapper



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

# 全局常量

from .i18n import ZH_CN


class GraphState:
    # 模块级全局变量
    _state = {
        'logdir': '',
        'current_tag': '',
        'current_run': '',
        'current_file_data': {},
        'current_hierarchy': {},
        'config_data': {
            "run": "",
            "npuUnMatchNodes": [],
            "benchUnMatchNodes": [],
            "npuMatchNodes": {},
            "benchMatchNodes": {},
            "manualMatchNodes": {},
            
        },
        'config_info': {},  # 全局的配置信息
        'first_run_tags': {},
        'runs': {},
        'update_precision_cache': {},  # {node_name{precision，...}}，方便查询精度，提高性能
        'all_node_info_cache': {},  # {rank_step_micro_step:{node_name:node_info,...}}，方便查询节点信息，提高性能
        'lang': ZH_CN,
    }

    @staticmethod
    def init_defaults():
        """
        初始化全局变量的默认值
        """
        GraphState._state = {
            'logdir': '',
            'current_tag': '',
            'current_run': '',
            'current_file_data': {},
            'current_hierarchy': {},
            'config_data': {
                "run": "",
                "npuUnMatchNodes": [],
                "benchUnMatchNodes": [],
                "npuMatchNodes": {},
                "benchMatchNodes": {},
                "manualMatchNodes": {},
                
            },
            'config_info': {},  # 全局的配置信息
            'first_run_tags': {},
            'runs': {},
            'update_precision_cache': {},
            'all_node_info_cache': {},
            'lang': ZH_CN,
        }

    @staticmethod
    def set_global_value(key, value):
        """
        设置全局变量的值
        """
        GraphState._state[key] = value

    @staticmethod
    def get_global_value(key, default=None):
        """
        获取全局变量的值
        """
        return GraphState._state.get(key, default)

    @staticmethod
    def reset_global_state():
        """
        重置所有全局变量为默认值
        """
        GraphState.init_defaults()




# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------


from msprobe.core.common.utils import CompareException
from msprobe.core.common.log import logger
from msprobe.core.compare.atb_data_compare import compare_atb_mode
from msprobe.core.compare.auto_compare import compare_auto_mode
from msprobe.core.compare.offline_data_compare import compare_offline_data_mode
from msprobe.core.compare.torchair_acc_cmp import compare_torchair_mode
from msprobe.infer.offline.compare.msquickcmp.main import compare_offline_model_mode


MODE_DISPATCHER = {
    'auto': compare_auto_mode,
    'offline_data': compare_offline_data_mode,
    'torchair': compare_torchair_mode,
    'offline_model': compare_offline_model_mode,
    'atb': compare_atb_mode
}

VALID_ARGS_MAP = {
    'auto': [
        '--mode', '-m', '--target_path', '-tp', '--golden_path', '-gp', '--output_path', '-o', 
        '--fuzzy_match', '-fm', '--cell_mapping', '-cm', '--api_mapping', '-am', '--data_mapping',
        '-dm', '--layer_mapping', '-lm', '--diff_analyze', '-da', '--rank', '--step',
        '--is_print_compare_log', '-tensor_log'
    ],
    'offline_data': [
        "-m", "-tp", "-gp", "-fr", "-qfr", "-cfr", "-o", "--mode", "--target_path", "--golden_path", 
        "--fusion_rule_file", "--quant_fusion_rule_file", "--close_fusion_rule_file", "--output_path"
    ],
    'torchair': [
        '--mode', '-m', '--target_path', '-tp', '--golden_path', '-gp', '--output_path', '-o'
    ],
    'offline_model': [
        '--mode', '-m', '--target_path', '-tp', '--golden_path', '-gp', '--output_path', '-o',
        '--rank', '--input_data', '--input_shape', '--output_size', '--dym_shape_range',
        '-ofs', '--onnx_fusion_switch'
    ],
    'atb': [
        '--mode', '-m', '--target_path', '-tp', '--golden_path', '-gp', '--output_path', '-o'
    ],
}


def check_valid_args(sys_argv, mode):
    valid_args = VALID_ARGS_MAP.get(mode)
    for arg in sys_argv:
        if arg.startswith("-") and arg not in valid_args:
            logger.error(f"Invalid argument '{arg}' for mode '{mode}'")
            raise CompareException(CompareException.INVALID_PARAM_ERROR)


def compare_cli(args, sys_argv):
    """
    Dispatch comparison based on mode parameter
    """
    mode = getattr(args, 'mode', 'auto')
    # Get the appropriate function based on mode
    compare_func = MODE_DISPATCHER.get(mode)
    if compare_func is None:
        logger.error(f"Invalid mode '{mode}'. Available modes: {list(MODE_DISPATCHER.keys())}")
        raise CompareException(CompareException.INVALID_COMPARE_MODE)
    check_valid_args(sys_argv, mode)
    # Execute the comparison function
    return compare_func(args)



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
gpu and cpu not implement benchmark function, supplementary benchmarking function implementation
"""

import os
from pkgutil import iter_modules
from importlib import import_module

package_path = os.path.dirname(os.path.realpath(__file__))
for _, module_name, _ in iter_modules([package_path]):
    module = import_module(f"{__name__}.{module_name}")
    for attr_name in dir(module):
        attr = getattr(module, attr_name)
        if callable(attr) and "npu_custom" not in attr_name:
            globals()[attr_name] = attr



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from msprobe.core.common.const import Const
from msprobe.core.common.exceptions import MsprobeException
from msprobe.pytorch.common.log import logger
from msprobe.pytorch.common.utils import is_torch_nn_module


class DebuggerConfig:
    def __init__(self, common_config, task_config, task, dump_path, level):
        self.dump_path = dump_path if dump_path else common_config.dump_path
        self.task = task or common_config.task or Const.STATISTICS
        self.rank = common_config.rank if common_config.rank else []
        self.step = common_config.step if common_config.step else []
        self.level = level or common_config.level or Const.LEVEL_L1
        self.scope = task_config.scope if task_config.scope else []
        self.list = task_config.list if task_config.list else []
        self.data_mode = task_config.data_mode if task_config.data_mode else ["all"]
        self.summary_mode = task_config.summary_mode if task_config.summary_mode else Const.STATISTICS
        self.framework = Const.PT_FRAMEWORK
        self.async_dump = common_config.async_dump if common_config.async_dump else False
        self.precision = common_config.precision if common_config.precision else Const.DUMP_PRECISION_LOW
        self.diff_nums = task_config.diff_nums if task_config.diff_nums else 1
        self.bench_path = getattr(task_config, "bench_path", None)
        self.check()
        self._check_statistics_config(task_config)

        if self.level == Const.LEVEL_L2:
            self.is_backward_kernel_dump = False
            self._check_and_adjust_config_with_l2()

    def check(self):
        if self.task and self.task not in Const.TORCH_TASK_LIST:
            raise MsprobeException(MsprobeException.INVALID_PARAM_ERROR,
                                   f"The task <{self.task}> is not in the {Const.TORCH_TASK_LIST}.")
        if self.level and self.level not in Const.LEVEL_LIST:
            raise MsprobeException(MsprobeException.INVALID_PARAM_ERROR,
                                   f"The level <{self.level}> is not in the {Const.LEVEL_LIST}.")
        if not self.dump_path:
            raise MsprobeException(MsprobeException.INVALID_PARAM_ERROR,
                                   f"The dump_path not found.")
        if not isinstance(self.async_dump, bool):
            raise MsprobeException(MsprobeException.INVALID_PARAM_ERROR,
                                   f"The parameters async_dump should be bool.")
        if self.task == Const.STRUCTURE and self.level not in [Const.LEVEL_L0, Const.LEVEL_MIX]:
            logger.warning_on_rank_0(
                f"When the task is set to structure, the level should be one of {[Const.LEVEL_L0, Const.LEVEL_MIX]}. "
                f"If not, the default level is {Const.LEVEL_MIX}."
            )
            self.level = Const.LEVEL_MIX
        if self.async_dump:
            if self.task == Const.TENSOR:
                if self.level == Const.LEVEL_DEBUG:
                    self.list = []  # async_dump + debug level case ignore list
                if not self.list and self.level != Const.LEVEL_DEBUG:
                    raise MsprobeException(
                        MsprobeException.INVALID_PARAM_ERROR,
                        f"The parameters async_dump is true in tensor task, the parameters list cannot be empty."
                    )
            if self.summary_mode == Const.MD5:
                raise MsprobeException(
                    MsprobeException.INVALID_PARAM_ERROR,
                    f"The parameters async_dump is true, the parameters summary_mode cannot be md5."
                )
        return True

    def check_model(self, models, token_range=None):
        if token_range and not models:
            error_info = "The 'model' parameter must be provided when token_range is not None"
            raise MsprobeException(MsprobeException.INVALID_PARAM_ERROR, error_info)

        if self.level not in [Const.LEVEL_L0, Const.LEVEL_MIX] and token_range is None:
            return

        if models is None:
            logger.error_on_rank_0(
                f"For level {self.level} or non-empty token_range, "
                f"PrecisionDebugger or start interface must receive a 'model' parameter.")
            raise MsprobeException(MsprobeException.INVALID_PARAM_ERROR, f"missing the parameter 'model'")

        if is_torch_nn_module(models):
            return

        if isinstance(models, (list, tuple)):
            error_model = None
            for model in models:
                if not is_torch_nn_module(model):
                    error_model = model
                    break
            if error_model is not None:
                error_info = (f"The 'model' parameter must be a torch.nn.Module or list[torch.nn.Module] "
                              f"type, currently there is an unsupported {type(error_model)} type.")
                raise MsprobeException(
                    MsprobeException.INVALID_PARAM_ERROR, error_info)
        else:
            error_info = (f"The 'model' parameter must be a torch.nn.Module or list[torch.nn.Module] "
                          f"type, currently there is an unsupported {type(models)} type.")
            raise MsprobeException(
                MsprobeException.INVALID_PARAM_ERROR, error_info)

    def _check_and_adjust_config_with_l2(self):
        if self.scope:
            raise MsprobeException(MsprobeException.INVALID_PARAM_ERROR,
                                   f"When level is set to L2, the scope cannot be configured.")
        if not self.list or len(self.list) != 1:
            raise MsprobeException(MsprobeException.INVALID_PARAM_ERROR,
                                   f"When level is set to L2, the list must be configured as a list with one api name.")
        if self.task != Const.TENSOR:
            raise MsprobeException(MsprobeException.INVALID_PARAM_ERROR,
                                   f"When level is set to L2, the task must be set to tensor.")

        api_name = self.list[0]
        if api_name.endswith(Const.BACKWARD):
            self.is_backward_kernel_dump = True
            api_forward_name = api_name[:-len(Const.BACKWARD)] + Const.FORWARD
            self.list.append(api_forward_name)

    def _check_statistics_config(self, task_config):
        if self.task != Const.STATISTICS:
            return
        self.tensor_list = []
        if not hasattr(task_config, "tensor_list"):
            return
        if self.level == Const.LEVEL_DEBUG and task_config.tensor_list:
            logger.warning_on_rank_0("When level is set to debug, the tensor_list will be invalid.")
            return
        self.tensor_list = task_config.tensor_list
